import momentsml
import matplotlib.pyplot as plt 
import numpy as np
import config
import os
import f2n


filepath = os.path.join(config.workdir, "psf.fits")

psf = momentsml.tools.io.fromfits(filepath)


print psf.shape

#plt.matshow(np.log10(psf), cmap=plt.get_cmap("gray"))
#plt.show()


myimage = f2n.fromfits(filepath)

#crop = 1
#myimage.crop(crop, -crop, crop, -crop)
    
myimage.setzscale(1.0e-5, 0.001)
myimage.makepilimage("log", negative = False)
#myimage.setzscale(-0.3, 1.0)
#myimage.makepilimage("lin", negative = False)
    
    
myimage.upsample(5)
pngimgpath = os.path.join(config.workdir, "psf.png")
myimage.tonet(pngimgpath)





import momentsml.sim
import momentsml.tools as tl
import numpy as np
import random # np.random.choice is only available for newer numpys...
import os
	

def trunc_gaussian(m, s, minval, maxval):
	"""
	A truncated Gaussian distrib
	"""
	assert maxval > minval
	a, b = (float(minval) - float(m)) / float(s), (float(maxval) - float(m)) / float(s)
	distr = scipy.stats.truncnorm(a, b, loc=m, scale=s)
	return distr.rvs()


def trunc_rayleigh(sigma, max_val):
	"""
	A truncated Rayleigh distribution
	"""
	assert max_val > sigma
	tmp = max_val + 1.0
	while tmp > max_val:
		tmp = np.random.rayleigh(sigma)
	return tmp


def contracted_rayleigh(sigma, max_val, p):
	"""
	A distribution with finite support.
	"""
	tmp = np.random.rayleigh(sigma)
	return (tmp / np.power(1 + np.power(tmp / max_val, p), 1.0 / p))


def psf_field_1(x, y):
	"""
	A simple PSF "model": this function returns PSF parameters as function of the (x, y) position
	in a plane defined by x and y in [0, 1].
	We also return the equivalent a, b, theta, which is only used for making plots.
	
	We use the definition "varepsilon" of the paper, equivalent to "g" in galsim. g = (a-b)/(a+b), and (a+b)/2 = sigma
	"""
	
	g_amp = 0.25 # maximum g
	default_sigma = 2.0 # average sigma
	sigma_amp = 0.25 # maximum deviation from default sigma
	
	tru_psf_g1 = g_amp * 2.0 * (x - 0.5)
	tru_psf_g2 = g_amp * y
	tru_psf_sigma = default_sigma + sigma_amp * ((x + y) - 1.0) 
	
	tru_psf_g = np.hypot(tru_psf_g1, tru_psf_g2)
	tru_psf_a = tru_psf_sigma * (1.0 + tru_psf_g)
	tru_psf_b = tru_psf_sigma * (1.0 - tru_psf_g)
	tru_psf_theta = np.arctan2(tru_psf_g2, tru_psf_g1) / 2.0
	
	return {"tru_psf_x":x, "tru_psf_y":y, 
		"tru_psf_sigma":tru_psf_sigma, "tru_psf_g1":tru_psf_g1, "tru_psf_g2":tru_psf_g2, "tru_psf_g":tru_psf_g,
		"tru_psf_a":tru_psf_a, "tru_psf_b":tru_psf_b, "tru_psf_theta":tru_psf_theta
		}
	


class Fiducial(momentsml.sim.params.Params):
	"""
	Fiducial parameters
	"""
	
	def __init__(self, name=None, snc_type=1, shear=0, noise_level=1.0, min_tru_rad=2.0, min_tru_sb=1.0, varpsf_type=None):
		"""
		- snc_type is the number of shape noise cancellation rotations
		- shear is the maximum shear to be drawn, 0 for no shear
		- noise_level 1.0 means that the fiducial noise level will be used, 0 means no noise
		"""
		
		momentsml.sim.params.Params.__init__(self)
		if name is not None:
			self.name = name
		self.snc_type = snc_type
		self.shear = shear
		self.noise_level = noise_level # So the default is Gaussian noise with std of 1.0
		self.min_tru_rad = min_tru_rad
		self.min_tru_sb = min_tru_sb
		self.varpsf_type = varpsf_type

	def draw_constants(self):
		"""
		Settings that don't change
		"""
		
		tru_type = 1 # Sersic
		snc_type = self.snc_type # The type of shape noise cancellation. 0 means none, n means n-fold
		tru_sky_level = 0.0 # GREAT3 has no Poisson noise, just flat Gaussian. # in ADU, just for generating noise, will not remain in the image
		tru_gain = -1.0 # in photons/ADU. Make this negative to have no Poisson noise
		tru_read_noise = self.noise_level
		tru_pixel = -1.0 # If positive, adds an extra convolution by that many pixels to the simulation process
		
		return {"snc_type":snc_type, "tru_type":tru_type, "tru_sky_level":tru_sky_level,
			"tru_gain":tru_gain, "tru_read_noise":tru_read_noise, "tru_pixel":tru_pixel,
			}

	def draw_s(self):
		"""
		Shear and magnification
		"""
		
		if self.shear > 0.0 and self.shear < 10.0:
			tru_s1 = np.random.uniform(-self.shear, self.shear)
			tru_s2 = np.random.uniform(-self.shear, self.shear)		
		else:
			tru_s1 = 0.0
			tru_s2 = 0.0
		
		tru_mu = 1.0
		return {"tru_s1":tru_s1, "tru_s2":tru_s2, "tru_mu":tru_mu}


	def draw_psf(self):
		"""
		Draws the parameters of the PSF
		"""
		if self.varpsf_type is None:
			tru_psf_sigma = 2.0 # -> FWHM = 4.7 pixels
			tru_psf_g1 = 0.0
			tru_psf_g2 = 0.0
			return {"tru_psf_sigma":tru_psf_sigma, "tru_psf_g1":tru_psf_g1, "tru_psf_g2":tru_psf_g2}
		else:
			# We draw a random position on the chip seen as plane with x and y in [0, 1]
			tru_psf_x = np.random.uniform(0.0, 1.0)
			tru_psf_y = np.random.uniform(0.0, 1.0)
			# From this position, we compute a g1, g2 and sigma of the PSF
			
			if self.varpsf_type is 1:
				return psf_field_1(tru_psf_x, tru_psf_y)
			else:
				raise NotImplemented()

	def stat(self):
		"""
		Called for each catalog (stat is for stationnary)
		"""
		return self.draw_constants()


	def draw(self, ix, iy, nx, ny):
		"""
		Called for each galaxy	
		"""
		
		tru_g = trunc_rayleigh(0.2, 0.6)
		tru_theta = 2.0 * np.pi * np.random.uniform(0.0, 1.0)		
		(tru_g1, tru_g2) = (tru_g * np.cos(2.0 * tru_theta), tru_g * np.sin(2.0 * tru_theta))
		
		tru_sersicn = random.choice(np.linspace(1, 4, 10))
		
		tru_sb = np.random.uniform(self.min_tru_sb, 15.0)
		tru_rad = np.random.uniform(self.min_tru_rad, 8.0)
		
		tru_flux = np.pi * tru_rad * tru_rad * tru_sb
			
		out = { # It's ugly to not directly fill this dict, but this makes it clearer what is actually returned:
			"tru_flux":tru_flux,
			"tru_rad":tru_rad,
			"tru_g1":tru_g1,
			"tru_g2":tru_g2,
			"tru_sersicn":tru_sersicn,
			"tru_g":tru_g,
			"tru_theta":tru_theta,
			"tru_sb": tru_sb,
			
		}
		
		out.update(self.draw_s()) # Here the shear gets drawn for each "galaxy"
		out.update(self.draw_psf()) # Idem, random for each galaxy
		out.update(self.draw_constants())
		
		return out
		


class Fiducial_statshear(Fiducial):
	"""
	To train weights for shear: constant shear per catalog
	"""
	
	def __init__(self, **kwargs):
		Fiducial.__init__(self, **kwargs)
		
	def stat(self):
		"""
		Supercedes what draw returns, called once for each catalog
		"""
		out = {}
		out.update(self.draw_s()) # Here, it's called for each catalog, and supercedes the galaxy values.
		out.update(self.draw_psf()) # Idem, one random PSF for each catalog
		out.update(self.draw_constants())
		
		return out
				



import matplotlib
matplotlib.use('svg')

from matplotlib import rc
#rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
# for Palatino and other serif fonts use:
#rc('font',**{'family':'serif','serif':['Palatino']})
#rc('text', usetex=True)

import matplotlib.pyplot as plt
plt.rcParams['svg.fonttype'] = 'none'

import momentsml.sim
import numpy as np
import random # np.random.choice is only available for newer numpys...
import scipy.stats
import config

import itertools
import galsim

import measfcts
import os
import glob

import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)


import f2n


cat = momentsml.tools.io.readpickle(os.path.join(config.simmeasdir, "figstamps", "groupmeascat.pkl"))

fitspaths = glob.glob(os.path.join(config.simdir, "figstamps", "*", "*.fits"))
assert len(fitspaths) == 1

fitspath = fitspaths[0]

print(fitspath)

#exit()


image_array = f2n.read_fits(fitspath)

cat["textx"] = cat["x"] - 12
cat["texty"] = cat["y"] - 9


import matplotlib.colors

print(cat["x", "y", "tru_mag", "snr"])


sf = f2n.SimpleFigure(image_array, z1=-5, z2=40, scale=3, withframe=False)
sf.draw() # norm=matplotlib.colors.LogNorm(vmin=-3, vmax=20))
sf.annotate(cat, x="textx", y="texty", text="S/N = {row[snr]:.1f}", color="white", fontsize=14) # Futher kwargs are passed to matplotib Text
#sf.show()
sf.save_to_file(os.path.join(config.workdir, "stamps.svg"))



"""
Custom and non-custom optimization algorithms to be used for training.
Functions take a train.Training object as first argument, perform their job, and return nothing.
They are almost methods of train.Training, but outsourced in this separate module.

"""

import numpy as np
import scipy.optimize


import logging
logger = logging.getLogger(__name__)



def bfgs(training, maxiter=100, gtol=1e-8, **kwargs):
	"""Calling scipy BFGS
	"""

	optres = scipy.optimize.fmin_bfgs(
			training.cost, training.params[training.paramslice],
			fprime=None,
			maxiter=maxiter, gtol=gtol,
			full_output=True, disp=False, retall=False, callback=training.callback, **kwargs)
		
	if len(optres) == 7:
		(xopt, fopt, gopt, Bopt, func_calls, grad_calls, warnflag) = optres
		training.cost(xopt) # To set the optimal parameters (even if this might not be needed, we don't want to depend on scipy internals...)
		logger.info("Done with optimization, {0} func_calls and {1} grad_calls".format(func_calls, grad_calls))
	else:
		logger.warning("Optimization output is fishy")


def multnetbfgs(training, nepochs=10, maxiter_sum=200, maxiter_mult=200, gtol=1e-8, **kwargs):
	"""A special version for MultNets, in development
	"""
	
	# We first start by optimizing only the sum layers, leaving the mult-layer as it is:
	if "maxiter" in kwargs:
		logger.warning("maxiter IN MULTNETBFGS IS FISHY. FOR NOW JUST REMOVING IT. IT WILL SOON BE RAISING AN ERROR")
		del kwargs["maxiter"]

	for epoch in range(nepochs):
		logger.info("Epoch {}/{} starting".format(epoch+1, nepochs))
		
		if maxiter_sum > 0:
			training.set_paramslice(mode="sum")
			bfgs(training, maxiter=maxiter_sum, gtol=gtol, **kwargs)
			logger.info("Optimisation sum layers done.")
		else:
			logger.info("Not optimising sum layers.")
	
		if maxiter_mult > 0:
			training.set_paramslice(mode="mult")
			bfgs(training, maxiter=maxiter_mult, gtol=gtol, **kwargs)
			logger.info("Optimisation mult layers done.")
		else:
			logger.info("Not optimising mult layers.")


def brute(training, maxiter=100, gtol=1e-6, **kwargs):
	"""Custom brute-force like optimization, for tests.
	"""
	logger.info("Starting brute optimization with options {}".format(kwargs))
	
	nparams = len(training.params)
	previouscost = None
	
	for iiter in range(maxiter):
		
		iniparams = training.params.copy()
		
		# We prepare a list for the params to be tested.
		paramslist = []
		costlist = []
		
		for iparam in range(nparams):
			for v in [-1., -0.51, -0.1, 0.1, 0.51, 1.0]:
				testparams = iniparams.copy()
				testparams[iparam] = v
				paramslist.append(testparams)
				cost = training.cost(testparams)
				costlist.append(cost)
				#print testparams, cost
			#for v in [-0.1, 0.1]:
			#	testparams = iniparams.copy()
			#	testparams[iparam] += v
			#	paramslist.append(testparams)
			#	cost = training.cost(testparams)
			#	costlist.append(cost)
			#	#print testparams, cost
				
		assert len(paramslist) == len(costlist)
		bestindex = np.argmin(costlist)
		bestparams = paramslist[bestindex].copy()
		bestcost = costlist[bestindex]
		
		training.cost(bestparams)
		training.callback(bestparams)	


		if previouscost == None:
			previouscost = bestcost
			continue # with next iteration
		
		if np.fabs((previouscost - bestcost) / bestcost) < gtol :
			#if np.allclose(bestparams, previousparams): # no, does not work because of zeros that make different params all have the same minimal cost.
			# Then nothing improved, we stop
			logger.info("Stopping this")
			break

		else:
			previouscost = bestcost




import matplotlib
matplotlib.use("AGG")

import os
import simparams
import measfcts
import f2n
import subprocess
import momentsml
import numpy as np

import logging

#logging.basicConfig(level=logging.INFO)
logging.basicConfig(format='PID %(process)06d | %(asctime)s | %(levelname)s: %(name)s(%(funcName)s): %(message)s',level=logging.INFO)



#### Parameters ###################

genworkdir = "/vol/fohlen11/fohlen11_1/mtewes/casereafig/"

name = "shearw"
#name = "shear"
#name = "ellip"

ncas = 5
nrea = 6
stampsize = 50

####################################




if name in ["shear", "ellip"]:
	sp = simparams.Simple1()
	sp.snc_type = nrea
elif name in ["shearw"]:
	sp = simparams.Simple2()
	sp.snc_type = 0
sp.sr = 0.0

workdir = os.path.join(genworkdir, name)
if not os.path.isdir(workdir):
	os.makedirs(workdir)
fitsimgpath = os.path.join(workdir, "img.fits")
catpath = os.path.join(workdir, "cat.pkl")
writecatpath = os.path.join(workdir, "writecat.txt")

nc = 1
n = ncas

if name is "shearw":
 	# We no longer use SNC for training weights!
	# nc = int(float(nrea)/2.0)
	# n = nc * ncas
	nc = nrea
	n = nrea * ncas

cat = momentsml.sim.stampgrid.drawcat(sp, n=n, nc=nc, stampsize=stampsize)

#exit()


if name == "ellip": # We give each rea the same ellipticity
	for i in range(ncas):
		g1 = cat[i*nrea]["tru_g1"]
		g2 = cat[i*nrea]["tru_g2"]
		print (g1, g2)
		for j in range(nrea):
			pass
			cat[i*nrea + j]["tru_g1"] = g1
			cat[i*nrea + j]["tru_g2"] = g2
			

if name == "shearw": # We give each case a different shear
	# No, in fact this looks weird. Better give zero or almost zero shear.
	for i in range(ncas):
		tru_s = 0.1
		tru_theta = 2.0 * np.pi * np.random.uniform(0.0, 1.0)
		(tru_s1, tru_s2) = (tru_s * np.cos(2.0 * tru_theta), tru_s * np.sin(2.0 * tru_theta))
		for j in range(nrea):
			cat[i*nrea + j]["tru_s1"] = tru_s1
			cat[i*nrea + j]["tru_s2"] = tru_s2

print cat["tru_g", "tru_g1", "tru_g2"]
print momentsml.tools.table.info(cat)
#exit()

momentsml.sim.stampgrid.drawimg(cat, simgalimgfilepath=fitsimgpath, simtrugalimgfilepath=None, simpsfimgfilepath=None)
#subprocess.Popen("ds9 {}".format(fitsimgpath), shell=True)


cat.meta["img"] = momentsml.tools.imageinfo.ImageInfo(
	fitsimgpath,
	xname="x",
	yname="y",
	stampsize=stampsize,
	pixelscale=1.0
	)

cat = momentsml.meas.galsim_adamom.measfct(cat, stampsize=stampsize, variant="wider")
cat = momentsml.meas.skystats.measfct(cat, stampsize=stampsize)
cat = momentsml.meas.snr.measfct(cat, gain=1.0)

if name == "shear" or name == "ellip":
	
	cat = momentsml.tools.table.groupreshape(cat, groupcolnames=["tru_flux"])
	momentsml.tools.table.addstats(cat, "snr")
	momentsml.tools.io.writepickle(cat, catpath)
	print momentsml.tools.table.info(cat)
	writecat = cat["tru_flux", "snr_mean"]
	print writecat
	writecat.write(writecatpath, format="ascii")

if name == "shearw":
	
	
	momentsml.tools.io.writepickle(cat, catpath)
	print momentsml.tools.table.info(cat)
	writecat = cat["x", "y", "snr"]
	print writecat
	writecat.write(writecatpath, format="ascii")



for icas in range(ncas):

	myimage = f2n.fromfits(fitsimgpath)
	myimage.crop(0, stampsize*nrea, icas*stampsize, (icas+1)*stampsize)
	
	#myimage.setzscale(-0.1, 3)
	#myimage.makepilimage("log", negative = False)
	myimage.setzscale(-0.3, 1.0)
	myimage.makepilimage("lin", negative = False)
	
	
	myimage.upsample(5)
	pngimgpath = os.path.join(workdir, "case_{}.png".format(icas))
	myimage.tonet(pngimgpath)

	subprocess.Popen("display {}".format(pngimgpath), shell=True)






"""
A MultNet is a variant of a Net, with a first hidden layer performing multiplications.
This first layer needs special initialization, and is not optimized at the same pace as the second part.


Things to keep in mind:
 - in case you call Net.setidentity, you WANT to call MultNet.multini again to set the initial params of the first mult layer.
 - you probably don't want to add noise to the mult layer if you're not optimizing it. One way of doing this is to use multwscale=0.0, multbscale=0.0.

"""

import numpy as np

import logging
logger = logging.getLogger(__name__)

from . import net
from . import layer
from . import utils
from . import act
from . import data

class MultNet(net.Net):


	def __init__(self, ni, nhs, mwlist=None, **kwargs):
		"""
		:param nhs: specify only the sum-hidden-layers, the mult-layer is inserted automatically!
		"""
		
		# MultNet specific tests:
		for n in nhs:
			if n < 0:
				raise RuntimeError("Would be fine for me, but is probably a mistake.")
	
		self.mwlist = mwlist
		nmult = ni + len(mwlist)
		net.Net.__init__(self, ni, [-nmult]+nhs, **kwargs)
	
		# Yes, it makes sense to directly set the weights of the first layer, as mwlist is given at this stage.
		self.multini()
		

	
	def multini(self):
		"""
		Initialize the mult layer in a very custom way.
		The first ni neurons are set to simply transport.
		The settings for the other ones are controlled by mwlist.
		
		:param mwlist: list of tuples containing the weights for each neuron of the mult-layer in addition
			to the "initial-identity" neurons.
			Example: mwlist=[(1, 1)] adds a single neuron that multiplies the first two inputs.
		
		"""
		
		logger.info("Initializing mult-layer with mwlist={}...".format(self.mwlist))
		
		if self.mwlist is None:
			raise RuntimeError("Weird: multini with mwlist None")
		
		#self.layers[0].setzero() # this will be done by setidentity.
		self.layers[0].setidentity(onlyn=self.ni)
		
		# And now the custom part, for the neurons which come after self.ni :
		for (i, item) in enumerate(self.mwlist):
			if len(item) > self.ni:
				raise RuntimeError("One item of your mwlist has too many weights!")
			for (inputi, weight) in enumerate(item):
				self.layers[0].weights[self.ni+i,inputi] = weight
			
		
		
		
	
	def get_paramslice(self, mode=None):
		"""
		Returns a slice object describing the parameters that should be optimized, in the context of a given mode.
		
		:param mode: If "sum", all parameters of the sum-layers are included.
			If "mult", only the parameters of the first mult-layer are in.
		
		"""
		
		logger.info("Preparing a paramslice to mode '{}'...".format(mode))
		
		ntotparams = self.nparams()
		nmultparams = self.layers[0].nparams()
		nsumparams = ntotparams - nmultparams
		
		if mode == "mult":
			# Those params, of the first layer, are the **last** in the paramslist, so we skip the first part:
			return slice(nsumparams, ntotparams)
		if mode == "sum":
			# We use only the first params:
			return slice(0, nsumparams)
		elif mode == None: # Empty slice, use all params
			return slice(None)	
		else:
			raise ValueError("Unknown mode!")
		
		



"""
Parameters describing how galaxies should be drawn
"""
import momentsml.sim
import numpy as np
import scipy.stats



class G3Sersics(momentsml.sim.params.Params):
	"""
	Sersic profiles for GREAT3 training, validation, and tests.
	"""

	def __init__(self, name=None, snc_type=1, shear=0, noise_level=1, distmode="G3", obstype="ground"):
		"""
		- snc_type is the number of shape noise cancellation rotations
		- shear is the maximum shear to be drawn, 0 for no shear
		- noise_level > 0 means that the subfields noise level will be used, 0 means no noise
		- distmode switches between different parameter distributions
		- obstype also affects parameter distributions to match the different G3 branches ("ground" or "space")
		"""
		momentsml.sim.params.Params.__init__(self)
		if name is not None:
			self.name = name
		self.snc_type = snc_type
		self.shear = shear
		self.noise_level = noise_level
		self.distmode = distmode
		self.obstype = obstype
		

	def draw_s(self):
		"""
		Shear and magnification
		"""
		if self.shear > 0:
			tru_s1 = np.random.uniform(-self.shear, self.shear)
			tru_s2 = np.random.uniform(-self.shear, self.shear)	
			
			#tru_s = 0.1
			#tru_theta = 2.0 * np.pi * np.random.uniform(0.0, 1.0)	
			#(tru_s1, tru_s2) = (tru_s * np.cos(2.0 * tru_theta), tru_s * np.sin(2.0 * tru_theta))
			
		else:
			tru_s1 = 0.0
			tru_s2 = 0.0
		
		tru_mu = 1.0
		return {"tru_s1":tru_s1, "tru_s2":tru_s2, "tru_mu":tru_mu}


	def draw_g(self):
		"""
		Galaxy ellipticity
		"""
		if self.distmode == "G3" or self.distmode == "train":
			# Simple distributions roughly tuned to ressemble GREAT3
			tru_g = contracted_rayleigh(0.2, 0.7, 4)
			tru_theta = 2.0 * np.pi * np.random.uniform(0.0, 1.0)		
			(tru_g1, tru_g2) = (tru_g * np.cos(2.0 * tru_theta), tru_g * np.sin(2.0 * tru_theta))
		
		elif self.distmode == "uni":
			# More "uniform" distribs for nicer filling in plots
			tru_g1 = np.random.uniform(-0.7, 0.7)
			tru_g2 = np.random.uniform(-0.7, 0.7)
			tru_g = np.hypot(tru_g1, tru_g2)
			tru_theta = 0.0 # Not needed, in fact.
			
		return {"tru_g1":tru_g1, "tru_g2":tru_g2, "tru_g":tru_g, "tru_theta":tru_theta}


	def draw_rad(self):
		"""
		Galaxy half-light-radius
		"""
		if self.distmode == "G3":

			if self.obstype == "ground":
				tru_rad = trunc_gaussian(1.0, 0.8, 0.75, 3.0)
			elif self.obstype == "space":
				#tru_rad = trunc_gaussian(3.0, 2.4, 2.25, 9.0) # Just a factor 3 larger than "ground", given that it's in pixels.
				tru_rad = trunc_gaussian(2.5, 3.5, 1.25, 10.0)
		
		elif self.distmode == "uni":
			
			if self.obstype == "ground":
				tru_rad = np.random.uniform(0.1, 3.0)
			elif self.obstype == "space":
				tru_rad = np.random.uniform(0.1, 10.0)
				
		elif self.distmode == "train": # Avoid the hopelessly small galaxies
		
			if self.obstype == "ground":
				tru_rad = np.random.uniform(0.75, 3.0)
			elif self.obstype == "space":
				tru_rad = np.random.uniform(1.25, 10.0)
		
		return {"tru_rad":tru_rad}
		
		
	def draw_sersicn(self, iy, ny):
		"""
		Galaxy Sersic index
		For GalSim performance it's better to not draw them completely randomly over the field.
		- iy: index of the y position in the field
		- ny: maximum index of the y position in the field
		"""
		if self.distmode == "G3": # Note the power 2, seems more representative
			
			if self.obstype == "ground":
				tru_sersicn =  0.5 + (float(iy)/float(ny))**2 * 2.0
			elif self.obstype == "space":
				tru_sersicn =  0.5 + (float(iy)/float(ny))**2 * 3.5
				
		elif self.distmode == "uni" or self.distmode == "train":
		
			tru_sersicn =  0.5 + (float(iy)/float(ny)) * 3.5
				
		return {"tru_sersicn":tru_sersicn}


	def draw_sb(self):
		"""
		Galaxy surface brightness
		Note that this is not used for GREAT3.
		"""
		
#		if self.distmode == "G3":
#			
#			if self.obstype == "ground":		
#				tru_sb = np.random.normal(1.5, 0.75)
#			elif self.obstype == "space":	
#				tru_sb = np.random.normal(0.2, 0.05)
#		
#		elif self.distmode == "uni" or self.distmode == "train":
#			
#			if self.obstype == "ground":				
#				tru_sb = np.random.uniform(0.8, 2.3)
#			elif self.obstype == "space":
#				tru_sb = np.random.uniform(0.15, 0.30)
		
		tru_sb = -1.0
		return {"tru_sb":tru_sb}


	def draw_flux(self):
		"""
		Galaxy flux
		"""
	
		if self.distmode == "G3":
			
			if self.obstype == "ground":
				tru_flux = trunc_gaussian(15, 20, 10, 200)		
				
			elif self.obstype == "space":	
				tru_flux = trunc_gaussian(0, 30, 10, 200)
		
		elif self.distmode == "uni":
		
			if self.obstype == "ground":				
				tru_flux = np.random.uniform(5, 80)
				
			elif self.obstype == "space":
				tru_flux = np.random.uniform(5, 80)
		
		
		elif self.distmode == "train":
			
			if self.obstype == "ground":				
				tru_flux = np.random.uniform(10, 100)
				
			elif self.obstype == "space":
				tru_flux = np.random.uniform(10, 100)
	
	
		return {"tru_flux":tru_flux}


	def draw_constants(self):
		"""
		Settings that don't change
		"""
		
		tru_type = 1 # 1 means Seric
		snc_type = self.snc_type # The type of shape noise cancellation. 0 means none, n means n-fold
		tru_sky_level = 0.0 # GREAT3 has no Poisson noise, just flat Gaussian. # in ADU, just for generating noise, will not remain in the image
		tru_gain = -1.0 # in photons/ADU. Make this negative to have no Poisson noise
		tru_read_noise = self.noise_level # in photons if gain > 0.0, otherwise in ADU. Set this to zero to have no flat Gaussian noise
		tru_pixel = -1.0 # If positive, adds an extra convolution by that many pixels to the simulation process

		return {"snc_type":snc_type, "tru_type":tru_type, "tru_sky_level":tru_sky_level,
			"tru_gain":tru_gain, "tru_read_noise":tru_read_noise, "tru_pixel":tru_pixel}


	def stat(self):
		"""
		Called for each catalog (stat is for stationnary)
		"""
			
		return self.draw_constants()
		


	def draw(self, ix, iy, nx, ny):
		"""
		Called for each galaxy	
		"""
	
		if self.distmode not in ["G3", "uni", "train"]:
			raise RuntimeError("Unknown distmode.")
		
		out = {}
		out.update(self.draw_s())
		out.update(self.draw_g())
		out.update(self.draw_rad())
		out.update(self.draw_sb())
		out.update(self.draw_flux())
		out.update(self.draw_sersicn(iy, ny))
		
		
		# We no longer use SB for GREAT3
		#out["tru_flux"] = 2.0 * np.pi * out["tru_rad"] * out["tru_rad"] * out["tru_sb"] # The 2 is from *half*-light-radius
				
		return out
		
		


class G3Sersics_statshear(G3Sersics):
	"""
	To train weights for shear: constant shear per catalog
	"""
	
	def __init__(self, **kwargs):
		G3Sersics.__init__(self, **kwargs)
		
	def stat(self):
		"""
		Supercedes what draw returns, called once for each catalog
		"""
		
		out = {}
		out.update(self.draw_s())
		out.update(self.draw_constants())
		return out
				
			
			

def trunc_gaussian(m, s, minval, maxval):
	"""
	A truncated Gaussian distrib
	"""
	assert maxval > minval
	a, b = (float(minval) - float(m)) / float(s), (float(maxval) - float(m)) / float(s)
	distr = scipy.stats.truncnorm(a, b, loc=m, scale=s)
	return distr.rvs()


def trun_rayleigh(sigma, max_val):
	"""
	A truncated Rayleigh distribution
	"""
	assert max_val > sigma
	tmp = max_val + 1.0
	while tmp > max_val:
		tmp = np.random.rayleigh(sigma)
	return tmp


def contracted_rayleigh(sigma, max_val, p):
	"""
	A distribution with finite support.
	"""
	tmp = np.random.rayleigh(sigma)
	return (tmp / np.power(1 + np.power(tmp / max_val, p), 1.0 / p))






import config
import os
import numpy as np
import momentsml.plot

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse, Arrow, FancyArrowPatch
import momentsml.plot.figures as figures
import matplotlib.ticker as ticker

import simparams

import logging
logger = logging.getLogger(__name__)


from matplotlib import rc
rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
## for Palatino and other serif fonts use:
#rc('font',**{'family':'serif','serif':['Palatino']})
rc('text', usetex=True)

fig = plt.figure(figsize=(8,4.0))

w = 0.3
cbarw = 0.02
h = 0.6

ax1 = fig.add_axes([0.1, 0.2, w, h])
ax2 = fig.add_axes([0.5, 0.2, w, h])
ax3 = fig.add_axes([0.82, 0.2, cbarw, h])


extra_space = 0.1

#ax1 = plt.subplot(1, 3, 1)
X1D = np.linspace(0.0, 1.0, 9)
Y1D = np.linspace(0.0, 1.0, 9)

for x in X1D:
	for y in Y1D:
	
		psf = simparams.psf_field_1(x, y)
		x = psf["tru_psf_x"]
		y = psf["tru_psf_y"]
		
		gscale = 0.55
		g1 = gscale * psf["tru_psf_g1"]
		g2 = gscale * psf["tru_psf_g2"]
		
		arrow = FancyArrowPatch(posA=(x-g1/2.0, y-g2/2.0), posB=(x+g1/2.0, y+g2/2.0), path=None, arrowstyle='-|>', mutation_scale=6, fc="black")
		ax1.add_artist(arrow)
		
		
#ax1.set_aspect('equal')
ax1.set_xlim([0-extra_space, 1+extra_space])
ax1.set_ylim([0-extra_space, 1+extra_space])

ax1.set_xlabel(r"$x$ position in field")
ax1.set_ylabel(r"$y$ position in field")

ax1.set_title(r"PSF field in ($\varepsilon_1$, $\varepsilon_2$)-space")



#ax2 = plt.subplot(1, 3, 2)

cmap = matplotlib.cm.get_cmap("plasma")

fwhmpersigma = 2.3548
minsigma = 1.75
maxsigma = 2.25

norm = matplotlib.colors.Normalize(vmin=minsigma * fwhmpersigma, vmax=maxsigma * fwhmpersigma)

X1D = np.linspace(0.0, 1.0, 9)
Y1D = np.linspace(0.0, 1.0, 9)
	
for x in X1D:
	for y in Y1D:
	
		psf = simparams.psf_field_1(x, y)
		x = psf["tru_psf_x"]
		y = psf["tru_psf_y"]
		
		escale = 0.03
		a = escale * psf["tru_psf_a"]
		b = escale * psf["tru_psf_b"]
		ellipse = Ellipse((x, y), a, b, np.rad2deg(psf["tru_psf_theta"]))		
		#ellipse.set_clip_box(ax2.bbox)
		ellipse.set_facecolor(cmap(norm(psf["tru_psf_sigma"] * fwhmpersigma)))
		ax2.add_artist(ellipse)


#ax3 = plt.subplot(1, 3, 3)
#pos3 = ax3.get_position()
#pos2 = ax2.get_position()

#ax3.set_position((pos2.x0 + pos2.width + 0.1, pos2.y0,  pos3.width / 20.0, pos2.height))


cbar = matplotlib.colorbar.ColorbarBase(ax3, cmap=cmap, norm=norm, orientation='vertical')
cbar.set_label(r'FWHM [pix]')
			
#ax2.set_aspect('equal')
ax2.set_xlim([0-extra_space, 1+extra_space])
ax2.set_ylim([0-extra_space, 1+extra_space])
ax2.set_xlabel(r"$x$ position in field")
ax2.set_ylabel(r"$y$ position in field")
ax2.set_title(r"PSF field in real space")

#plt.tight_layout()
plt.savefig(os.path.join(config.valdir, "psf_field.pdf"))
plt.show()






import momentsml.sim
import numpy as np
import random # np.random.choice is only available for newer numpys...

import itertools





class Simple1(momentsml.sim.params.Params):
	"""
	No PSF, just plain Gaussians and Gaussian noise
	"""
	
	
	def __init__(self):
		momentsml.sim.params.Params.__init__(self)

	
	def stat(self):
		"""
		stat: called for each catalog (stat is for stationnary)
		"""
		return {"snc_type":1}
	
	
	def draw(self, ix, iy, nx, ny):
		"""
		draw: called for each galaxy	
		"""
	
		########## Noise ##########

		tru_sky_level = 0.0
		tru_gain = -1.0
		tru_read_noise = 1.0
		
		#########  No Lensing
		
		tru_s1 = 0.0
		tru_s2 = 0.0
		tru_mu = 1.0

		
		########## Galaxy ##########
		
		tru_type = 0 # 0 Gaussian, 1 sersic	
		
		#minflux = 100.0
		#maxflux = 1000.0
		#tru_flux = minflux + (maxflux - minflux) * float(ny - iy)/float(ny)
		#tru_flux = [100, 150, 200][iy]
		
		
		refrad = (4.3/2.0)
		#minrad = refrad/2.0
		#maxrad = refrad * 2.0
		#tru_rad = minrad + (maxrad-minrad) * float(ix)/float(nx)
		rads = np.array([0.75*refrad, refrad, 1.5*refrad])
		tru_rad = rads[ix]
		
		fact = 0.54
		fluxes = np.array([rads * flux * fact  for flux in [50, 100, 150, 200]]).transpose()
		#print fluxes.shape
		#exit()
		#fluxes = np.array([100, 150, 200])
		
		tru_flux = fluxes[ix, iy]
		
		
		tru_sigma = tru_rad / 1.1774 # We take Croppers "extension of the source" as the half-light-diameter
		
		
		tru_g = 0.0
		tru_theta = 0.0	
		(tru_g1, tru_g2) = (tru_g * np.cos(2.0 * tru_theta), tru_g * np.sin(2.0 * tru_theta))
		
		tru_g = np.hypot(tru_g1, tru_g2)
			
		
		
		return { # It's ugly to not directly fill this dict, but this makes it clearer what is actually returned:
					
			"tru_type":tru_type,
			"tru_flux":tru_flux,
			"tru_rad":tru_rad,
			"tru_sigma":tru_sigma,
			#"tru_sersicn":tru_sersicn,
			"tru_g1":tru_g1,
			"tru_g2":tru_g2,
			"tru_g":tru_g, # only useful for some plots
			
			"tru_sky_level":tru_sky_level, # in ADU, just for generating noise, will not remain in the image
			"tru_gain":tru_gain, # in photons/ADU. Make this negative to have no Poisson noise
			"tru_read_noise":tru_read_noise, # in photons if gain > 0.0, otherwise in ADU.Set this to zero to have no flat Gaussian noise
			
			"tru_s1":tru_s1,
			"tru_s2":tru_s2,
			"tru_mu":tru_mu,
					
		}





"""

2D plots of predictions versus truth, but shown in bins of truth, as points with error bars.



"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.ticker import MaxNLocator
from matplotlib.ticker import AutoMinorLocator
from matplotlib.lines import Line2D

from .. import tools
from . import utils

import astropy

import logging
logger = logging.getLogger(__name__)





def res(ax, cat, feattru, featpre, featc=None, nbins=10, ncbins=3, ebarmode="bias", showidline=True, metrics=True, equalcount=False):
	"""
	Shows residues of a predicted feature featpre in bins of the corresponding truth feattru.
	If featc is specified (c stands for color), this analysis is done in several bins of featc
	(say to check that the shear estimate is unbiased for different galaxy sizes...).
	
	:param feattru: Feature with "true" value, for the x-axis
	:param featpre: Feature with "predicted" value. Residuals will be computed comparing this to feattru.
		Trick: the low and high of featpre are used to define the plot range for the residues.
	:param featc: Feature to make color-bins
	:param nbins: number of bins in the x-axis
	:param ncbins: number of color-bins
	:ebarmode: either "bias" or "scatter". If "bias", errorbars show the std on the mean, and not just the std of the full scatter.
	
	"""

	logger.info("Preparing res plot of '{}' in bins of '{}'".format(featpre.colname, feattru.colname))
	
	errorbarkwargs = {"ls":"None", "marker":".", "lw":1.5, "mew":1.5}

	# First, we get the 1D data, removing all masked points.
	features = [feattru, featpre]
	if featc is not None:
		features.append(featc)
	data = tools.feature.get1Ddata(cat, features, keepmasked=False)
	
	# We compute the residues, for every point.
	assert "res" not in data.colnames
	data["res"] = data[featpre.colname] - data[feattru.colname]
	
	# Defining the bins of the x axis:
	xbinrange = utils.getrange(data, feattru)
			
	# And for the "color" axis:
	if featc is not None:
		logger.info("Building bins in {}...".format(featc.colname))
		cbinrange = utils.getrange(data, featc)
		
		if equalcount is True:
			# With the latest numpy this should work without using a list comprehension, but ok...
			cbinlims = np.array([np.percentile(data[featc.colname], q) for q in np.linspace(0.0, 100.0, ncbins+1)])
		else:
			cbinlims = np.linspace(cbinrange[0], cbinrange[1], ncbins+1)
		
		cbinlows = cbinlims[0:-1]
		cbinhighs = cbinlims[1:]
		cbincenters = 0.5 * (cbinlows + cbinhighs)
		assert len(cbincenters) == ncbins
		
		if ncbins == 1:
			coloriter = iter(["black"])
		elif ncbins <= 4:
			coloriter = iter(["red", "darkgreen", "blue", "magenta"])
		else:
			coloriter=iter(matplotlib.cm.brg(np.linspace(0,1,ncbins)))
		offsetscale = 0.5*((xbinrange[1] - xbinrange[0])/float(nbins))/float(ncbins)

		for i in range(ncbins):
			
			color = next(coloriter)
			label = "{0:.2g} < {1} < {2:.2g}".format(cbinlows[i], featc.nicename, cbinhighs[i])
			offset = (i - float(ncbins)/2) * offsetscale
			
			# We build the subset of data that is in this color bin:
			selcbin = tools.table.Selector(label, [("in", featc.colname, cbinlows[i], cbinhighs[i])])
			cbindata = selcbin.select(data)
			if len(cbindata) == 0:
				continue
			cbinfrac = float(len(cbindata)) / float(len(data))
			label += " ({0:.0%})".format(cbinfrac)
			
			# And we perform the line regression
			if metrics is True:
				try:
					md = tools.metrics.metrics(cbindata,
						tools.feature.Feature(feattru.colname), # Redefining those to get rid of any rea settings that don't apply to cbindata
						tools.feature.Feature("res"),
						pre_is_res=True)
					
					metricstext = "m = %.1f +/- %.1f, c = %.1f +/- %.1f" % (md["m"]*1000.0, md["merr"]*1000.0, md["c"]*1000.0, md["cerr"]*1000.0)				
					label += "\n" + metricstext
					ax.plot(np.array(xbinrange), md["m"]*np.array(xbinrange)+md["c"], color=color, ls="-")
					
				except:
					logger.warning("Metrics compuation failed", exc_info = True)
			
			
			# And now bin this in x:
			cbinsumma = utils.summabin(cbindata[feattru.colname], cbindata["res"], xbinrange=xbinrange, nbins=nbins, equalcount=False)
			
			yerr = cbinsumma["ystds"]
			#yerr = np.array([cbinsumma["ylowps"], cbinsumma["yhighps"]])
			yerrbias = yerr / np.sqrt(cbinsumma["ns"])
	
			if ebarmode == "scatter":
				ax.errorbar(cbinsumma["xbincents"]+offset, cbinsumma["ymeans"], yerr=yerr, color=color, label=label, **errorbarkwargs)
			elif ebarmode == "bias":
				ax.errorbar(cbinsumma["xbincents"]+offset, cbinsumma["ymeans"], yerr=yerrbias, color=color, label=label, **errorbarkwargs)
	
			
		
			

	else:	
	
		color="black"
		binsumma = utils.summabin(data[feattru.colname], data["res"], xbinrange=xbinrange, nbins=nbins)
	
		yerr = binsumma["ystds"]
		#yerr = np.array([bindata["ylowps"], bindata["yhighps"]])
		yerrbias = yerr / np.sqrt(binsumma["ns"])
	
		if ebarmode == "scatter":
			ax.errorbar(binsumma["xbincents"], binsumma["ymeans"], yerr=yerr, color=color, **errorbarkwargs)
		elif ebarmode == "bias":
			ax.errorbar(binsumma["xbincents"], binsumma["ymeans"], yerr=yerrbias, color=color, **errorbarkwargs)
	
		if metrics is True:
				try:
					md = tools.metrics.metrics(data,
						tools.feature.Feature(feattru.colname), # Redefining those to get rid of any rea settings that don't apply to cbindata
						tools.feature.Feature("res"),
						pre_is_res=True)
					
					metricstext = "m = %.1f +/- %.1f, c = %.1f +/- %.1f" % (md["m"]*1000.0, md["merr"]*1000.0, md["c"]*1000.0, md["cerr"]*1000.0)				
					ax.plot(np.array(xbinrange), md["m"]*np.array(xbinrange)+md["c"], color=color, ls="-")
					ax.annotate(metricstext, xy=(0.0, 1.0), xycoords='axes fraction', xytext=(8, -12), textcoords='offset points', ha='left', va='top')
			
				except:
					logger.warning("Metrics compuation failed", exc_info = True)
			
	
	
		
	if showidline: # Show the identity line
		idlinekwargs = {"ls":"--", "color":"gray", "lw":1}
		ax.plot(xbinrange, (0.0, 0.0), **idlinekwargs)

	# We want minor ticks:
	ax.xaxis.set_minor_locator(AutoMinorLocator(5))
	ax.yaxis.set_minor_locator(AutoMinorLocator(5))
	
	ax.set_ylabel(featpre.nicename + " - " + feattru.nicename)
	ax.set_ylim(featpre.low, featpre.high)
	ax.set_xlabel(feattru.nicename)
	ax.set_xlim(xbinrange)

	if featc:
		ax.legend(prop={"size":"small"})
	
	
	
	
	
	
	
	
	
	
	
	
"""	
	binlims = np.linspace(binrange[0], binrange[1], nbins+1)
	bincenters = 0.5 * (binlims[0:-1] + binlims[1:])
	assert len(bincenters) == nbins
	
	binindices = np.digitize(data[feattru.colname], bins=binlims)
	inrangeindices = np.arange(nbins)+1 # those indices that are within the binlims
	assert len(inrangeindices) == nbins
	
	
	# Defining the bins in "colour"
	if featc is not None:
		cbinrange = utils.getrange(data, featc)
		cbinlims = np.linspace(cbinrange[0], cbinrange[1], ncbins+1)
		cbincenters = 0.5 * (cbinlims[0:-1] + cbinlims[1:])
		assert len(cbincenters) == ncbins
		
	
	ymeans = []
	ystds = []
	ylowps = []
	yhighps = []
	ns = []
		
	for ind in inrangeindices: # We loop over the bins
		
		inbools = binindices == ind # a boolean array
		nin = np.sum(inbools)
		
		if nin < 2:
			ymeans.append(np.nan)
			ystds.append(np.nan)
			ylowps.append(np.nan)
			yhighps.append(np.nan)
			ns.append(np.nan)
			continue
		
		ns.append(nin)
		
		thesexvals = data[feattru.colname][inbools]
		theseyvals = data["res"][inbools]
		
		#assert len(thesexvals) == nin
		#assert len(theseyvals) == nin
		
		ymeans.append(np.mean(theseyvals))
		ystds.append(np.std(theseyvals))
		
		ylowps.append(np.fabs(np.percentile(theseyvals, 15.8) - np.mean(theseyvals)))
		yhighps.append(np.fabs(np.percentile(theseyvals, 84.1) - np.mean(theseyvals)))
		
		
				
	#errorbarkwargs = {"capthick":0, "zorder":-100}
	errorbarkwargs = {"color":"black", "ls":"None", "marker":".", "lw":1.0, "mew":1.0}
	
	#yerr = ystds
	yerr = np.array([ylowps, yhighps])
	yerr_bias = yerr / np.sqrt(np.array(ns))
	yerr_stdbias = np.array(ystds) / np.sqrt(np.array(ns))
	
	if ebarmode == "scatter":
		ax.errorbar(bincenters, ymeans, yerr=yerr, **errorbarkwargs)
	elif ebarmode == "bias":
		ax.errorbar(bincenters, ymeans, yerr=yerr_bias, **errorbarkwargs)
	
	#yerr = np.array(ystds) / np.sqrt(np.array(ns))
	#ax.errorbar(bincenters, ymeans, yerr=yerr, color="red", ls="None", marker=".")
	
	if showidline: # Show the identity line
		idlinekwargs = {"ls":"--", "color":"gray", "lw":1}
		ax.plot(binrange, (0.0, 0.0), **idlinekwargs)


	# We want minor ticks:
	ax.xaxis.set_minor_locator(AutoMinorLocator(5))
	ax.yaxis.set_minor_locator(AutoMinorLocator(5))
	
	ax.set_ylabel(featpre.nicename + " - " + feattru.nicename)
	ax.set_ylim(featpre.low, featpre.high)
	ax.set_xlabel(feattru.nicename)
	ax.set_xlim(binrange)
	
	
"""	




















def bin(ax, cat, featx, featy, nbins=10, selector=None, showselector=True, title=None, showidline=True, metrics=False,
	ebarmode="scatter", yisres=True):
	"""
	Summarizes featy in bins of featx.
	
	:param ebarmode: controls what should be shown as error-bars
	:param yisres: if True, featy will be treated as residues when computing the metrics, and idline will be horizontal.
		Note that for a bin-plot, this is the default.
	
	"""

	logger.info("Preparing bin plot of '{}' versus '{}'".format(featy.colname, featx.colname))
	
	nall = len(cat)
	nsel = nall
	if selector:
		cat = selector.select(cat)
		nsel = len(cat)
		selfrac = float(nsel)/float(nall)
	
	features = [featx, featy]
	data = tools.feature.get1Ddata(cat, features, keepmasked=False)
	
	# Finding the range to bin:
	if featx.low is not None and featx.high is not None: 
		binrange = (featx.low, featx.high)
	else:
		binrange = (np.min(data[featx.colname]), np.max(data[featx.colname]))

	logger.info("Binning data along '{name}' in {nbins} bins in the range {ran}...".format(name=featx.colname, nbins=nbins, ran=binrange))

	binlims = np.linspace(binrange[0], binrange[1], nbins+1)
	bincenters = 0.5 * (binlims[0:-1] + binlims[1:])
	assert len(bincenters) == nbins
	
	binindices = np.digitize(data[featx.colname], bins=binlims)
	
	inrangeindices = np.arange(nbins)+1 # the indices that are within the binlims
	assert len(inrangeindices) == nbins
	
	
	ymeans = []
	ystds = []
	ylowps = []
	yhighps = []
	ns = []
		
	for ind in inrangeindices: # We loop over the bins
		
		inbools = binindices == ind # a boolean array
		nin = np.sum(inbools)
		
		if nin < 2:
			ymeans.append(np.nan)
			ystds.append(np.nan)
			ylowps.append(np.nan)
			yhighps.append(np.nan)
			ns.append(np.nan)
			continue
		
		ns.append(nin)
		
		thesexvals = data[featx.colname][inbools]
		theseyvals = data[featy.colname][inbools]
		
		#assert len(thesexvals) == nin
		#assert len(theseyvals) == nin
		
		ymeans.append(np.mean(theseyvals))
		ystds.append(np.std(theseyvals))
		
		ylowps.append(np.fabs(np.percentile(theseyvals, 15.8) - np.mean(theseyvals)))
		yhighps.append(np.fabs(np.percentile(theseyvals, 84.1) - np.mean(theseyvals)))
		
	assert len(ymeans) == len(bincenters)
	assert len(ylowps) == len(bincenters)
	assert len(yhighps) == len(bincenters)
				
	#errorbarkwargs = {"capthick":0, "zorder":-100}
	errorbarkwargs = {"color":"black", "ls":"None", "marker":".", "lw":1.0, "mew":1.0}
	
	#yerr = ystds
	yerr = np.array([ylowps, yhighps])
	yerr_bias = yerr / np.sqrt(np.array(ns))
	yerr_stdbias = np.array(ystds) / np.sqrt(np.array(ns))
	
	if ebarmode == "scatter":
		ax.errorbar(bincenters, ymeans, yerr=yerr, **errorbarkwargs)
	elif ebarmode == "bias":
		ax.errorbar(bincenters, ymeans, yerr=yerr_bias, **errorbarkwargs)
	
	#yerr = np.array(ystds) / np.sqrt(np.array(ns))
	#ax.errorbar(bincenters, ymeans, yerr=yerr, color="red", ls="None", marker=".")
	
	
	if showidline: # Show the identity line
		idlinekwargs = {"ls":"--", "color":"gray", "lw":1}
		ax.plot(binrange, (0.0, 0.0), **idlinekwargs)

	if metrics:
		
		logger.info("Now computing metrics for this res plot...")
		try:
			metrics = tools.metrics.metrics(cat, featx, featy, pre_is_res=yisres)
			
			metrics_text = "predfrac: %.3f\nRMSD: %.5f\nm*1e3: %.1f +/- %.1f\nc*1e3: %.1f +/- %.1f" % (metrics["predfrac"], metrics["rmsd"], metrics["m"]*1000.0, metrics["merr"]*1000.0, metrics["c"]*1000.0, metrics["cerr"]*1000.0)
			#metrics_text = "RMSD = %.5f\nm = %.1f +/- %.1f, c = %.1f +/- %.1f e-3" % (metrics["rmsd"], metrics["m"]*1000.0, metrics["merr"]*1000.0, metrics["c"]*1000.0, metrics["cerr"]*1000.0)
			
			
			ax.annotate(metrics_text, xy=(0.0, 1.0), xycoords='axes fraction', xytext=(8, -22), textcoords='offset points', ha='left', va='top')
		except:
			logger.warning("Metrics compuation failed", exc_info = True)

	if selector and showselector:
		utils.showselector(ax, selector)


	# We want minor ticks:
	ax.xaxis.set_minor_locator(AutoMinorLocator(5))
	ax.yaxis.set_minor_locator(AutoMinorLocator(5))
	
	ax.set_ylabel(featy.nicename)
	ax.set_ylim(featy.low, featy.high)
	ax.set_xlabel(featx.nicename)
	ax.set_xlim(binrange)

	if selector and showselector:
		ax.set_title(selector.name + " ({:3.0f} %)".format(100.0*selfrac))

	if title:
		ax.set_title(title)

	



"""
2-D histogram like plots using hexagonal bins
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.ticker import MaxNLocator
from matplotlib.ticker import AutoMinorLocator
from matplotlib.lines import Line2D


from .. import tools

import logging
logger = logging.getLogger(__name__)



def hexbin(ax, cat, featx, featy, featc=None, makecolorbar=True, cblabel="Counts", title=None, text=None, showidline=False, idlinekwargs=None, **kwargs):
	"""
	Plots hexbin tiles on the axes.
	
	By default, if featc is None, the tile colors represent number counts,
	thus making it a 2-D histogram.
	If you pass a featc, note that you can also specify, via the kwarg reduce_C_function,
	what statistic to compute in each tile.
	
	:param ax: a matplotlib.axes.Axes object
	:param cat: an astropy table 
	:param featx: a Feature object telling what to draw on the x axis
	:param featy: idem for y
	:param featc: a Feature to use for the tile colormap.
	:param makecolorbar: set to False to avoid making the colorbar.
		This can be useful in particular when you want to overplot something else on the same axes.
	:param cblabel: what to write as label for the colorbar if featc is not specified.
		You might want to use "log(counts)" if usign the kwarg bins="log"
	:param title: the title to place on top of the axis.
	:param text: some text to be written in the figure (top left corner)
		As we frequently want to do this, here is a simple way to do it.
		For more complicated things, add the text yourself to the axes.
	
	Commonly used kwargs:
	
	- gridsize = 15
	- mincnt = 30
	- cmap = "Blues" # add "_r" to reverse... (one of my favs for residuals is "rainbow", "gnuplot2",)
	- reduce_C_function = np.mean (default), tools.calc.rmsd_delta (to compute RMSD from prediction errors)
	- bins = "log" (works only for "counts", so without a featc. Otherwise, specify the reduce_C_function...
		
	More info at http://matplotlib.org/api/axes_api.html?highlight=hexbin#matplotlib.axes.Axes.hexbin
	
	How to select good colormaps: http://matplotlib.org/users/colormaps.html
	"""
	
	if featc == None:
		featctxt = "counts"
	else:
		featctxt = featc.colname
	logger.info("Preparing hexbin plot of '%s' in the ('%s', '%s') plane" % (featctxt, featx.colname, featy.colname))
	
	# Getting the data (without masked points):
	features = [featx, featy]
	if featc is not None:
		features.append(featc)
	data = tools.feature.get1Ddata(cat, features, keepmasked=False)		
	
	if featx.low is not None and featx.high is not None and featy.low is not None and featy.high is not None: 
		extent = (featx.low, featx.high, featy.low, featy.high) # xmin xmax ymin ymax
	else:
		extent = None
	
	# Some good default values:
	mykwargs = {"gridsize":30, "cmap":"gist_heat_r", "extent":extent}
	
	# We overwrite these mykwargs with any user-specified kwargs:
	mykwargs.update(kwargs)
	
	if featc == None: # We use number counts to determine color
		cbstuff = ax.hexbin(data[featx.colname], data[featy.colname], **mykwargs)
		
	else: # We use featc to determine color
		cbstuff = ax.hexbin(data[featx.colname], data[featy.colname], C=data[featc.colname], vmin=featc.low, vmax=featc.high,  **mykwargs)
		
	if makecolorbar:
		divider = make_axes_locatable(ax)
		cax = divider.append_axes("right", "5%", pad="3%")	
		cb = plt.colorbar(cbstuff, cax)
	
		if featc == None:
			cb.set_label(cblabel)
		else:
			cb.set_label(featc.nicename)

	if showidline: # Show the "diagonal" identity line
	
		# It would be nice to get this working with less code
		# (usign get_lims and axes transforms, for instance)
		# But in the meantime this seems to work fine.
		# It has to be so complicated to keep the automatic ranges working if low and high are None !
		
		# For "low":
		if featx.low is None or featy.low is None: # We use the data...
			minid = max(np.min(data[featx.colname]), np.min(data[featy.colname]))
		else:
			minid = max(featx.low, featy.low)
		# Same for "high":
		if featx.high is None or featy.high is None: # We use the data...
			maxid = min(np.max(data[featx.colname]), np.max(data[featy.colname]))
		else:
			maxid = min(featx.high, featy.high)
			
		if idlinekwargs == None:
			idlinekwargs = {}
		myidlinekwargs = {"ls":"--", "color":"black", "lw":1}
		myidlinekwargs.update(idlinekwargs)	
		
		# And we plot the line:
		ax.plot((minid, maxid), (minid, maxid), **myidlinekwargs)

	# The usual axes limit setting :
	ax.set_xlim(featx.low, featx.high)
	ax.set_ylim(featy.low, featy.high)
	ax.set_xlabel(featx.nicename)
	ax.set_ylabel(featy.nicename)
	# And the minor ticks:
	ax.xaxis.set_minor_locator(AutoMinorLocator(5))
	ax.yaxis.set_minor_locator(AutoMinorLocator(5))

	# Writing any text:
	if text:
		ax.annotate(text, xy=(0.0, 1.0), xycoords='axes fraction', xytext=(8, -8), textcoords='offset points', ha='left', va='top')
	if title:
		ax.set_title(title)




import f2n
import numpy as np
import os

import logging

#logging.basicConfig(level=logging.INFO)
logging.basicConfig(format='PID %(process)06d | %(asctime)s | %(levelname)s: %(name)s(%(funcName)s): %(message)s',level=logging.INFO)





fitsimgpath = "/vol/fohlen11/fohlen11_1/mtewes/snr_fig/v1/img.fits"

pngimgpath = fitsimgpath + ".png"

myimage = f2n.fromfits(fitsimgpath)
myimage.setzscale(-2.0, 10.0)
myimage.makepilimage("lin", negative = False)
	
	
myimage.upsample(5)
myimage.tonet(pngimgpath)









import momentsml
import momentsml.learn
import os
import numpy as np

import logging
loggerformat='\033[1;31m%(levelname)s\033[1;0m: %(name)s(%(funcName)s): \033[1;21m%(message)s\033[1;0m'
#loggerformat='PID %(process)06d | %(asctime)s | %(levelname)s: %(name)s(%(funcName)s): %(message)s'
logging.basicConfig(format=loggerformat,level=logging.INFO)

workdir = "/vol/fohlen12/data1/mtewes/MomentsML/papereuclidlike"
gemsdir = "/vol/fohlen12/data1/mtewes/GEMS"
spectrumpath = "/vol/fohlen12/data1/mtewes/Pickles_spectra/pickles_uk_27.fits"


simdir = os.path.join(workdir, "sim")
simmeasdir = os.path.join(workdir, "simmeas")
traindir = os.path.join(workdir, "train")
valdir = os.path.join(workdir, "val")


sourcecat = os.path.join(workdir, "cat.pkl")

stampsize = 64 # used for measuring
drawstampsize = stampsize # Used for drawing

psfoversampling = 5
psfstampsize = 64


datasets = {
	#"tp":"tp-1-e",
	#"tp":"tp-05-e",
	
	
	"tp":"tp-1-pretrain", # <-------
	#"tp":"tp-05-pretrain",
	
	
	#"vp":"vp-1",
	#"vp":"vp-1s",
	"vp":"vp-1-gems",
	
	#"tw":"tw-1", #<-------
	#"tw":"tw-1-snr-above-10",
	#"tw":"tw-1-snr-above-10-sigma-above-1.5", #<------- old version

	"tw":"tw-2-snr-above-10-sigma-above-1.5", # <------ this is now the best it seems
	#"tw":"tw-2-snr-above-10-sigma-above-1.5-mbfrac1", # <-- running, seems to be equally good
	
	# SNC stuff:	# No, not needed!
	#"tw":"tw-1-snc",
	#"tw":"tw-1-snc-snr-above-10",
	#"tw":"tw-1-snc-snr-above-10-sigma-above-1.5",
	
	
	
	#"vo":"vo-1",
	#"vo":"vo-1-snr-above-10",
	#"vo":"vo-1-snr-above-10-sigma-above-1.5",

	#"vo":"vo-2",
	#"vo":"vo-2-snr-above-10-sigma-above-1.5",
	#"vo":"tw-2-snr-above-10-sigma-above-1.5",
	
	#"vo":"vo-3",
	"vo":"vo-3-snr-above-10-sigma-above-1.5", # <------ 
	#"vo":"vo-3-realcut",
	
	"si":"si-1-gems",
	#"si":"si-1-uni",
	
}


shearconflist = [
	
	("mlconfig/ada5s1.cfg", "mlconfig/sum55.cfg"), # <--- We use those
	("mlconfig/ada5s2.cfg", "mlconfig/sum55.cfg")
	
	
]

weightconflist = [
	("mlconfig/ada5s1w.cfg", "mlconfig/sum5w.cfg"),
	("mlconfig/ada5s2w.cfg", "mlconfig/sum5w.cfg")

]


#######################################################################################################################################

# Some automatically generated names, used for saving figures etc depending on your above settings.
# Not meant to be changed by the user...

sconfname = os.path.splitext(os.path.basename(shearconflist[0][1]))[0] # extracts e.g. "sum55"
valname = "{}_with_{}_on_{}".format(datasets["tp"], sconfname, datasets["vp"])
wconfname = os.path.splitext(os.path.basename(weightconflist[0][1]))[0] # extracts e.g. "sum55w"
wvalname = "{}_and_{}_with_{}_{}_on_{}".format(datasets["tp"], datasets["tw"], sconfname, wconfname, datasets["vo"])

for d in [simdir, traindir, valdir]:
	if not os.path.exists(d):
		os.makedirs(d)



"""

Plot of m and c as datapoints versus bins of some feature.



"""

import numpy as np
import matplotlib.pyplot as plt
#import matplotlib.cm
#from mpl_toolkits.axes_grid1 import make_axes_locatable
#from matplotlib.ticker import MaxNLocator
#from matplotlib.ticker import AutoMinorLocator
#from matplotlib.lines import Line2D
import matplotlib.ticker as ticker

from .. import tools
from . import utils

import astropy

import logging
logger = logging.getLogger(__name__)




def mcbin(ax, cat, feattru, featpre, featbin, featprew=None, nbins=10, binlims=None, showbins=True, comp=0, showlegend=False, regressmethod=1, sigma_shape=0.2):
	"""
	Makes plots of mu and c bias terms evaluated in bins of some (usually true) parameter.
	
	feattru: the true parameter (for example tru_s1)
	featpre: the predicted point estimate (for example pre_s1)
	featbin: the feature to bin in (for example tru_g)
	featprew: the predicted weights (optional, for example pre_s1w)

	comp: 0, 1 or 2, if not 0 selects the symbol and color to use
	
	Either give nbins (to get equal-part-of-population bins) or binlims (an array of the "n+1" bin limits to use).
	
	About the regressmethod:
	This affects how the weights given by featprew are used: if you don't have weights, you can ignore regressmethod.
	
	regressmethod = 1 : a weighted least squares regression is performed on *all* individual realizations within a bin.
		For this, a value of sigma_shape is required to turn the weights into shear errors and give meaningful error bars.
		
	regressmethod = 2 : the weights are used to compute the weighted average true shear within each *case*, and then a non-weighted
		least square regression is performed on the true-shear vs weighted-average-shear relation.
		In this case sigma_shape is not used, as it is not required.
		A pracitcal advantage of regessmethod 2 is that it gives meaningful errors even if SNC is used within the cases.
		You want to use this if your care about all cases equally, for instance if different cases have different PSFs.
	
	"""
	
	logger.info("Plot mcbin with featbin '{}'".format(featbin.colname))
	
	if binlims is None:		
		binrange = utils.getrange(cat, featbin)
		binlims = np.array([np.percentile(cat[featbin.colname], q) for q in np.linspace(0.0, 100.0, nbins+1)])
	else:
		nbins = len(binlims) - 1
		if nbins < 1:
			raise ValueError("Provide more binlims!")
	
	binlows = binlims[0:-1]
	binhighs = binlims[1:]
	#logger.info("binlims: {}".format(binlims))
		
	binpointcenters = []
	ms = []
	merrs = []
	cs = []
	cerrs = []

	for i in range(nbins):
			
		selbin = tools.table.Selector(featbin.colname, [("in", featbin.colname, binlows[i], binhighs[i])])
		bindata = selbin.select(cat)
		
		if regressmethod is 1 or featprew is None:
	
			# And we perform the linear regression
			# Redefining features, to get rid of any rea settings that don't apply here (?)
		
			if featprew is None:
				md = tools.metrics.metrics(bindata,
						feattru, 
						featpre,
						pre_is_res=False)
			else:
				md = tools.metrics.metricsw(bindata,
						feattru, 
						featpre,
						featprew,
						sigma_shape=sigma_shape
						)
		elif regressmethod is 2:
		
			# We first compute (weighted) average shear per case, for each selected case (meaningful for cases that have the same shear).
			# Then we perform a non-weighted regression to get m and c 
			# Advantage is for the weights: we don't care about sigma_shape or any SNC which modifies the shape noise
			
			
			if bindata[feattru.colname].shape[0] != len(bindata):
				raise RuntimeError("Using regressmethod 2 implies that the cases should all have the same feattru value!")
			
			tools.table.addstats(bindata, featpre.colname, wcol=featprew.colname)
			#print tools.table.info(bindata)
			
			wmeancolname = featpre.colname + "_wmean"
			featwmean = tools.feature.Feature(wmeancolname)
			
			md = tools.metrics.metrics(bindata,
						feattru, 
						featwmean,
						pre_is_res=False)
			
			
			
		ms.append(md["m"])
		merrs.append(md["merr"])
		cs.append(md["c"])
		cerrs.append(md["cerr"])
			
		binpointcenters.append(np.mean(bindata[featbin.colname]))
	
			
	
	if comp is 0:
		mkwargs = {}
		ckwargs = {}
	elif comp is 1:
		mkwargs = {"marker":'s', "color":"black", "label":r"$\mu_{}$ ".format(comp)}
		ckwargs = {"marker":'*', "color":"black", "label":r"$c_{}$ ".format(comp), "ls":":"}
	elif comp is 2:
		mkwargs = {"marker":'d', "color":"red", "label":r"$\mu_{}$ ".format(comp)}
		ckwargs = {"marker":'x', "color":"red", "label":r"$c_{}$ ".format(comp), "ls":":"}
	else:
		raise ValueError("Unknown comp")
	
	
	ax.errorbar(binpointcenters, ms, yerr=merrs, **mkwargs)
	ax.errorbar(binpointcenters, cs, yerr=cerrs, **ckwargs)
		
	if showbins:
		for x in binlims:
			ax.axvline(x, color='gray', lw=0.5)
	
	if showlegend:
		#plt.legend(loc="best", handletextpad=0.07, fontsize="small", framealpha=1.0, columnspacing=0.1, ncol=2)
		plt.legend(loc="best",  handletextpad=0.2, framealpha=1.0, columnspacing=0.1, ncol=2)



def make_symlog(ax, featbin, linthresh=2e-3, lim=1e-1):
	"""
	Converts the y axis to a symlog scale with custom ticks, usually for the mcbin-plot from above
	"""
	
	ax.set_yscale('symlog', linthreshy=linthresh)
	ax.set_ylim([-lim, lim])
	ticks = np.concatenate([np.arange(-linthresh, linthresh, 1e-3)])#, np.arange(lintresh, 1e-2, 9)])
	s = ax.yaxis._scale
	ax.yaxis.set_minor_locator(ticker.SymmetricalLogLocator(s, subs=[1., 2.,3.,4.,5.,6.,7.,8.,9.,-2.,-3.,-4.,-5.,-6.,-7.,-8.,-9.]))
	ticks = np.concatenate([ticks, ax.yaxis.get_minor_locator().tick_values(-.1, .1)])
	ax.yaxis.set_minor_locator(ticker.FixedLocator(ticks))
	
	
	xlim = (featbin.low, featbin.high)
	ax.fill_between(xlim, -linthresh, linthresh, alpha=0.2, facecolor='darkgrey')
	ax.set_xlim(xlim)




"""
Training z = x * y with a single product unit, testing this mult neuron.

Single realziation, no noise in the data, not inverse regression.

"""

import numpy as np
import tenbilac

import logging
logging.basicConfig(level=logging.INFO)


nrea = 10
ncas = 500

# We prepare the data:

# Range of x and y
xs = np.random.uniform(-5, 5, ncas)
ys = np.random.uniform(-10, 30, ncas)

zs = xs * ys 

inputs = np.array([xs, ys])
inputs = np.tile(inputs, (nrea, 1, 1))
# We add noise:
inputs += 0.001*np.random.randn(inputs.size).reshape(inputs.shape)

# This is 3D (rea, features=2, case)
#print inputs
#print inputs.shape


targets = np.array([zs])

# This is 2D (feature=1, case)
#print targets
#print targets.shape


inputnormer = tenbilac.data.Normer(inputs, type="sa1")
inputs = inputnormer(inputs)
targetnormer = tenbilac.data.Normer(targets, type="sa1")
targets = targetnormer(targets)


dat = tenbilac.data.Traindata(inputs=inputs, targets=targets)

net = tenbilac.multnet.MultNet(ni=2, mwlist=[(1, 1)], nhs=[3], no=1, actfctname="tanh", oactfctname="iden", multactfctname="iden", inames=["x", "y"], onames=["z"])

training = tenbilac.train.Training(net, dat, errfctname="msb", autoplot=True, autoplotdirpath="test_opt")


net.setidentity()
net.multini()

# We can even add noise afterwards, as long as we keep the mult-scales at zero:
net.addnoise(multwscale=0.0, multbscale=0.0, wscale=0.1, bscale=0.1)

tenbilac.plot.netviz(net, title="Ready!")

training.opt(algo="multnetbfgs", mbsize=None, mbfrac=1.0, mbloops=5, gtol=1e-10, nepochs=10, maxiter_sum=200, maxiter_mult=200)
#training.set_paramslice(mode="sum")
#training.opt(algo="bfgs", mbsize=None, mbfrac=1.0, mbloops=5, gtol=1e-10, maxiter=200)

print net.report()

tenbilac.plot.netviz(net, title="Done!")


outs = targetnormer.denorm(net.predict(inputs))
#outs = net.predict(inputs)
# Shape is (rea, neuron, case)

assert zs.size == ncas
residues = outs[:,0,:] - zs # We have only one neuron
# Shape is (rea, case)

# Stats for each case, over the realizations:
meanres = np.mean(residues, axis=0)
stdres = np.std(residues, axis=0)
assert meanres.size == ncas


#training.save("test.pkl", keepdata=True)
#exit()
import matplotlib.pyplot as plt


fig = plt.figure(figsize=(12, 5))

plt.subplot(1, 2, 1)
plt.scatter(xs, ys, c=meanres, s=80, marker="o", edgecolors="face")
plt.xlabel("x")
plt.ylabel("y")
plt.colorbar()
plt.title("Mean residue")

plt.subplot(1, 2, 2)
plt.scatter(xs, ys, c=stdres, s=80, marker="o", edgecolors="face")
plt.xlabel("x")
plt.ylabel("y")
plt.colorbar()
plt.title("Std residue")

plt.tight_layout()


plt.show()




"""
Error functions
These should work both on unmasked and on masked arrays "as expected".

The typical call is fct(predictions, targets, auxinputs), where auxinputs can be left out if the error function does not use any auxinputs.
"""

import numpy as np

import logging
logger = logging.getLogger(__name__)


def mse(predictions, targets, auxinputs=None):
	"""
	Standard MSE (mean square error), simply treats multiple realizations as if they were independent cases	
	
	:param predictions: 2D array (neuron, case) or 3D array (realization, neuron, case)
	:param targets: 2D array (neuron, case)

	"""
	
	# This same code works for 2D or 3D predictions:
	return np.mean(np.square(predictions - targets))

	
	
def msb(predictions, targets, auxinputs=None):
	"""
	Mean square bias
	
	:param predictions: 3D array (realization, neuron, case), should be appropiratedly masked (thus not directly the output of the net)
	:param targets: 2D array (neuron, case)
	
	"""
	
	if predictions.ndim == 3:
	
		biases = np.mean(predictions, axis=0) - targets # This is 2D, (label, case)
		return np.mean(np.square(biases))
	
	else:
		raise ValueError("Wrong pred shape")




def mab(predictions, targets, auxinputs=None):
	"""
	Mean absolute bias
	
	:param predictions: 3D array (realization, neuron, case), should be appropiratedly masked (thus not directly the output of the net)
	:param targets: 2D array (neuron, case)
	
	"""
	
	if predictions.ndim == 3:
	
		biases = np.mean(predictions, axis=0) - targets # This is 2D, (label, case)
		return np.mean(np.fabs(biases))
	
	else:
		raise ValueError("Wrong pred shape")


def msbwsignorm(predictions, targets, auxinputs):
	"""
	Weighted mean square bias of the auxinputs, where weights are obtained form predictions (use a sigmoid last layer!),
	and the weighting is normalized.
	
	We do NOT normalize the weighted average auxinputs
	"""
	
	if auxinputs is None:
		raise RuntimeError("This error function needs auxinputs.")
	
	assert predictions.ndim == 3
	assert auxinputs.ndim == 3
	assert targets.ndim == 2
	
	nt = targets.shape[0] # the number of targets = number of "predicted weights" = number of aux inputs (per case)
	assert auxinputs.shape[1] == nt
	assert predictions.shape[1] == nt 
	
	weights = predictions
	biases = np.mean(auxinputs * weights, axis=0) / np.mean(weights, axis=0) - targets # The mean is done along realizations, so this is 2D, (label, case)
			
	return np.mean(np.square(biases)) #+ np.square(np.max(predictions) - 1.0)



def mabwsignorm(predictions, targets, auxinputs):
	"""
	Same as msbwsignorm, but with L1 norm.
	"""
	
	if auxinputs is None:
		raise RuntimeError("This error function needs auxinputs.")
	
	assert predictions.ndim == 3
	assert auxinputs.ndim == 3
	assert targets.ndim == 2
	
	nt = targets.shape[0] # the number of targets = number of "predicted weights" = number of aux inputs (per case)
	assert auxinputs.shape[1] == nt
	assert predictions.shape[1] == nt 
	
	weights = predictions
	biases = np.mean(auxinputs * weights, axis=0) / np.mean(weights, axis=0) - targets # The mean is done along realizations, so this is 2D, (label, case)
			
	return np.mean(np.fabs(biases)) #+ np.square(np.max(predictions) - 1.0)




########## Below this line, things get experimental and might not make sense


	
def msrb(predictions, targets, auxinputs=None, rawterms=False):
	"""
	Mean square relative bias
	
	:param predictions: 3D array (realization, neuron, case), should be appropiratedly masked (thus not directly the output of the net)
	:param targets: 2D array (neuron, case)
	
	:param rawterms: if True, returns the "RB" of "MSRB" as (potentially masked) 2D array.
	
	"""
	
	if predictions.ndim == 3:
	
		#assert type(predictions) == np.ma.MaskedArray
		# Just as a test that this was not forgotten, no real need here.
		# No need, for this, for sure it also works without masks...
		
		biases = np.mean(predictions, axis=0) - targets # This is 2D, (label, case) : masked, but probably all masks are False.
		stds = np.std(predictions, axis=0) # idem
		
		if type(predictions) == np.ma.MaskedArray:
			reacounts = predictions.shape[0] - np.sum(predictions.mask, axis=0) + 0.0 # Number of realizations, 0.0 makes this floats # This is 2D (label, case)
		elif type(predictions) == np.ndarray:
			reacounts = predictions.shape[0] + 0.0
		else:
			raise RuntimeError("Type error in predictions.")
		
		errsonbiases = stds / np.sqrt(reacounts)
		relativebiases = biases / errsonbiases
		
		if rawterms:
			#return (biases, errsonbiases)
			return relativebiases # 2D (label, case), masked, but probably all masks are False.
		else:
			return np.mean(np.square(relativebiases))
		
	
	else:
		raise ValueError("Wrong pred shape")
	
	


def msre(predictions, targets, auxinputs=None):
	"""
	MSE with normalization by the scatter along the realizations (as MSRB is for MSB)

	Mean Square Error / Variance - 1

	"""

	if predictions.ndim == 3:
		
		stds = np.std(predictions, axis=0) # 2D, (label, case)

		return np.mean(np.square((predictions - targets) / stds) - 1.0) # The minus one helps to viz it on a log scale...
	
	else:
		raise ValueError("Wrong pred shape")




def msbwnet(predictions, targets, auxinputs=None):
	"""
	Mean square bias with weights, for training a WNet.
	This is the first errorfunction of a new type, it compares the weighted average of the predictions with the targets.
	
	There should be twice more prediction "neurons" than targets. The second half of the predictions is interpreted as weights
	
	:param predictions: 3D array (realization, neuron, case), should be appropiratedly masked (thus not directly the output of the net)
	:param targets: 2D array (neuron, case)
	
	"""
	
	if predictions.ndim == 3:
	
		nt = targets.shape[0] # the number of targets = number of "predicted parameters" = number of weights for these outputs
		assert predictions.shape[1] == 2 * nt # Indeed, these are the predicted parameters and the corresponding weights.
		
		predparams = predictions[:,:nt,:]
		predweights = 10**predictions[:,nt:,:]
		assert predparams.shape == predweights.shape
	
		biases = np.mean(predparams * predweights, axis=0) - targets # The mean is done along realizations, so this is 2D, (label, case)
	
		return np.mean(np.square(biases))
	
	else:
		raise ValueError("Wrong pred shape")
	


def msbpw(predictions, targets, auxinputs):
	"""
	with pre-weights
	"""
	
	assert predictions.ndim == 3
	assert auxinputs.ndim == 3
	assert targets.ndim == 2
	
	nt = targets.shape[0] # the number of targets = number of "predicted weights" = number of aux inputs (per case)
	assert auxinputs.shape[1] == nt
	assert predictions.shape[1] == nt 
	
	biases = np.mean(auxinputs * predictions, axis=0) / np.mean(auxinputs, axis=0) - targets # The mean is done along realizations, so this is 2D, (label, case)
	
	return np.mean(np.square(biases))


def msbw(predictions, targets, auxinputs):
	"""
	This errorfct is for Nets predicting weights only. It expresses the mean square weighted bias of the auxinputs
	
	The predictions should be masked, auxinputs as well. Targets are usually not masked.
	"""
	
	#logger.warning("Normalizes the weight, probably a bad idea !")
	assert predictions.ndim == 3
	assert auxinputs.ndim == 3
	assert targets.ndim == 2
	
	nt = targets.shape[0] # the number of targets = number of "predicted weights" = number of aux inputs (per case)
	assert auxinputs.shape[1] == nt
	assert predictions.shape[1] == nt 
	
	predweights = 10**predictions
	biases = np.mean(auxinputs * predweights, axis=0) / np.mean(predweights, axis=0) - targets # The mean is done along realizations, so this is 2D, (label, case)
		
	return np.mean(np.square(biases))
	
	
def msbwnonorm(predictions, targets, auxinputs):
	"""
	That normalization was a very bad idea...
	"""
	
	assert predictions.ndim == 3
	assert auxinputs.ndim == 3
	assert targets.ndim == 2
	
	nt = targets.shape[0] # the number of targets = number of "predicted weights" = number of aux inputs (per case)
	assert auxinputs.shape[1] == nt
	assert predictions.shape[1] == nt 
	
	predweights = 10**predictions
	biases = np.mean(auxinputs * predweights, axis=0) - targets # The mean is done along realizations, so this is 2D, (label, case)
		
	return np.mean(np.square(biases))
	
	
def msbwclip(predictions, targets, auxinputs):
	"""
	Direct, but limiting the weights to be positive.
	"""
	
	assert predictions.ndim == 3
	assert auxinputs.ndim == 3
	assert targets.ndim == 2
	
	nt = targets.shape[0] # the number of targets = number of "predicted weights" = number of aux inputs (per case)
	assert auxinputs.shape[1] == nt
	assert predictions.shape[1] == nt 
	
	predweights = np.clip(predictions, 1e-6, 1e15) # need large uppper limit, otherwise optimizer stops
	biases = np.mean(auxinputs * predweights, axis=0) - targets # The mean is done along realizations, so this is 2D, (label, case)
		
	return np.mean(np.square(biases))


def msbwsquare(predictions, targets, auxinputs):
	"""
	"""
	
	assert predictions.ndim == 3
	assert auxinputs.ndim == 3
	assert targets.ndim == 2
	
	nt = targets.shape[0] # the number of targets = number of "predicted weights" = number of aux inputs (per case)
	assert auxinputs.shape[1] == nt
	assert predictions.shape[1] == nt 
	
	predweights = predictions**2
	biases = np.mean(auxinputs * predweights, axis=0) - targets # The mean is done along realizations, so this is 2D, (label, case)
		
	return np.mean(np.square(biases))


def msbwinvsquare(predictions, targets, auxinputs):
	"""
	"""
	
	assert predictions.ndim == 3
	assert auxinputs.ndim == 3
	assert targets.ndim == 2
	
	nt = targets.shape[0] # the number of targets = number of "predicted weights" = number of aux inputs (per case)
	assert auxinputs.shape[1] == nt
	assert predictions.shape[1] == nt 
	
	predweights = 1.0/predictions**2
	biases = np.mean(auxinputs * predweights, axis=0) - targets # The mean is done along realizations, so this is 2D, (label, case)
		
	return np.mean(np.square(biases))



def msbwsig(predictions, targets, auxinputs):
	"""
	Similar as msbw, but we do not use this dangerous power of 10, and we do not normalize the weights at training
	
	This is for networks whose ouput is from a sigmoid layer, so positive by construction (and between 0 and 1).
	Weights are 1 * output.
	We do NOT normalize the weighted average auxinputs
	"""
	
	assert predictions.ndim == 3
	assert auxinputs.ndim == 3
	assert targets.ndim == 2
	
	nt = targets.shape[0] # the number of targets = number of "predicted weights" = number of aux inputs (per case)
	assert auxinputs.shape[1] == nt
	assert predictions.shape[1] == nt 
	
	biases = np.mean(auxinputs * 1.0*predictions, axis=0) - targets # The mean is done along realizations, so this is 2D, (label, case)
		
	return np.mean(np.square(biases))
	

def msbwsigtwo(predictions, targets, auxinputs):
	"""
	Similar as msbw, but we do not use this dangerous power of 10, and we do not normalize the weights at training
	
	This is for networks whose ouput is from a sigmoid layer, so positive by construction (and between 0 and 1).
	Weights are 2 * output.
	We do NOT normalize the weighted average auxinputs
	"""
	
	assert predictions.ndim == 3
	assert auxinputs.ndim == 3
	assert targets.ndim == 2
	
	nt = targets.shape[0] # the number of targets = number of "predicted weights" = number of aux inputs (per case)
	assert auxinputs.shape[1] == nt
	assert predictions.shape[1] == nt 
	
	biases = np.mean(auxinputs * 2.0*predictions, axis=0) - targets # The mean is done along realizations, so this is 2D, (label, case)
		
	return np.mean(np.square(biases))
	
def msbwsigten(predictions, targets, auxinputs):
	"""
	Similar as msbw, but we do not use this dangerous power of 10, and we do not normalize the weights at training
	
	This is for networks whose ouput is from a sigmoid layer, so positive by construction (and between 0 and 1).
	Weights are 10 * output.
	We do NOT normalize the weighted average auxinputs
	"""
	
	assert predictions.ndim == 3
	assert auxinputs.ndim == 3
	assert targets.ndim == 2
	
	nt = targets.shape[0] # the number of targets = number of "predicted weights" = number of aux inputs (per case)
	assert auxinputs.shape[1] == nt
	assert predictions.shape[1] == nt 
	
	biases = np.mean(auxinputs * 10.0*predictions, axis=0) - targets # The mean is done along realizations, so this is 2D, (label, case)
		
	return np.mean(np.square(biases))

def msbwsigthree(predictions, targets, auxinputs):
	"""
	Similar as msbw, but we do not use this dangerous power of 10, and we do not normalize the weights at training
	
	This is for networks whose ouput is from a sigmoid layer, so positive by construction (and between 0 and 1).
	Weights are 3 * output.
	We do NOT normalize the weighted average auxinputs
	"""
	
	assert predictions.ndim == 3
	assert auxinputs.ndim == 3
	assert targets.ndim == 2
	
	nt = targets.shape[0] # the number of targets = number of "predicted weights" = number of aux inputs (per case)
	assert auxinputs.shape[1] == nt
	assert predictions.shape[1] == nt 
	
	biases = np.mean(auxinputs * 3.0*predictions, axis=0) - targets # The mean is done along realizations, so this is 2D, (label, case)
		
	return np.mean(np.square(biases))



"""
Plots the figure of the paper 
"""

import config
import measfcts
import os
import numpy as np
import matplotlib.ticker as ticker
import matplotlib.colors
import matplotlib

import momentsml.plot
from momentsml.tools.feature import Feature
import matplotlib.pyplot as plt

import logging
logger = logging.getLogger(__name__)



#momentsml.plot.figures.set_fancy(14)
from matplotlib import rc
rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
## for Palatino and other serif fonts use:
#rc('font',**{'family':'serif','serif':['Palatino']})
rc('text', usetex=True)



valname = config.wvalname

#selstr = "snr-above-10"
selstr = "snr-above-10-sigma-above-1.5"
#selstr = "realcut"


if selstr in config.datasets["vo"] and not selstr in config.datasets["tw"]:
	#title=r"Retaining only galaxies with S/N $>$ 10.0 for the validation"
	#title=r"Retaining only galaxies with S/N $>$ 10.0 and {\tt adamom\_sigma} $>$ 1.5 for the validation"
	title=r"Same shear estimator applied on a selection of sources with S/N $>$ 10.0 and {\tt adamom\_sigma} $>$ 1.5 pixel:"
	if selstr == "realcut":
		title = r"Same shear estimator applied on a (theoretical) selection of sources in R and mag to match."
	
	mode = 2
	
elif selstr in config.datasets["vo"] and selstr in config.datasets["tw"]:
	#title=r"Retaining galaxies with S/N $>$ 10.0 for the training and the validation"
	#title=r"Retaining only galaxies with S/N $>$ 10.0 and {\tt adamom\_sigma} $>$ 1.5 for the training and the validation"
	#title=r"Using weight predictions trained with the selection S/N $>$ 10.0 and {\tt adamom\_sigma} $>$ 1.5 pixel:"
	title=r"Applying the same selection S/N $>$ 10.0 and {\tt adamom\_sigma} $>$ 1.5 pixel for both training and validation:"
	
	mode = 3

else:
	title=r"Without specific selection of training or validation galaxies:"
	mode = 1

valcatpath = os.path.join(config.valdir, valname + ".pkl")
cat = momentsml.tools.io.readpickle(valcatpath)


# Important so that masked point don't show up in the third panel
cat["snr"].mask = cat["pre_s1"].mask

#print momentsml.tools.table.info(cat)

for comp in ["1","2"]:

	
	#cat["pre_s{}w_norm".format(comp)] = cat["pre_s{}w".format(comp)] / np.max(cat["pre_s{}w".format(comp)])
	#momentsml.tools.table.addrmsd(cat, "pre_s{}".format(comp), "tru_s{}".format(comp))
	
	momentsml.tools.table.addstats(cat, "pre_s{}".format(comp), "pre_s{}w".format(comp))
	cat["pre_s{}_wbias".format(comp)] = cat["pre_s{}_wmean".format(comp)] - cat["tru_s{}".format(comp)]


resr = 0.01
wplotrea = -10

snr = Feature("snr", 3.0, 150.0, nicename="S/N", rea=wplotrea)
tru_mag = Feature("tru_mag", 20, 25.7, nicename="Magnitude", rea=wplotrea)
tru_rad = Feature("tru_rad", 0.0, 13.0, nicename=r"Half-light radius $R$ [pix]", rea=wplotrea)


tru_s1 = Feature("tru_s1", nicename=r"$g_1^{\mathrm{true}}$")
tru_s2 = Feature("tru_s2", nicename=r"$g_2^{\mathrm{true}}$")

pre_s1_bias = Feature("pre_s1_bias", -resr, resr, nicename=r"$\langle \hat{g}_{1} \rangle - g_{1}^{\mathrm{true}} $")
pre_s2_bias = Feature("pre_s2_bias", -resr, resr, nicename=r"$\langle \hat{g}_{2} \rangle - g_{2}^{\mathrm{true}} $")
pre_s1_wbias = Feature("pre_s1_wbias", -resr, resr, nicename=r"$\left(\sum\hat{g}_1 w_1 / \sum w_1 \right) - g_{1}^{\mathrm{true}}$")
pre_s2_wbias = Feature("pre_s2_wbias", -resr, resr, nicename=r"$\left(\sum\hat{g}_2 w_2 / \sum w_2 \right) - g_{2}^{\mathrm{true}}$")



def addmetrics(ax, xfeat, yfeat):
	metrics = momentsml.tools.metrics.metrics(cat, xfeat, yfeat, pre_is_res=True)
	line1 = r"$10^3 \mu=%.1f \pm %.1f $" % (metrics["m"]*1000.0, metrics["merr"]*1000.0)
	line2 = r"$10^3 c=%.2f \pm %.2f $" % (metrics["c"]*1000.0, metrics["cerr"]*1000.0)
	
	ax.annotate(line1, xy=(0.0, 1.0), xycoords='axes fraction', xytext=(8, -9), textcoords='offset points', ha='left', va='top', fontsize=12)
	ax.annotate(line2, xy=(0.0, 1.0), xycoords='axes fraction', xytext=(8, -22), textcoords='offset points', ha='left', va='top', fontsize=12)
	

fig = plt.figure(figsize=(10, 3))
plt.subplots_adjust(
	left  = 0.09,  # the left side of the subplots of the figure
	right = 0.93,    # the right side of the subplots of the figure
	bottom = 0.18,   # the bottom of the subplots of the figure
	top = 0.85,      # the top of the subplots of the figure
	wspace = 0.45,   # the amount of width reserved for blank space between subplots,
	                # expressed as a fraction of the average axis width
	hspace = 0.25,   # the amount of height reserved for white space between subplots,
					# expressed as a fraction of the average axis heightbottom=0.1, right=0.8, top=0.9)
	)

idlinekwargs = {"color":"black", "ls":"-"}

#==================================================================================================



ax = fig.add_subplot(1, 3, 1)
momentsml.plot.scatter.scatter(ax, cat, tru_s1, pre_s1_wbias, showidline=True, idlinekwargs=idlinekwargs, yisres=True)
addmetrics(ax, tru_s1, pre_s1_wbias)

#ax.set_xlabel("")
#ax.set_xticklabels([])
#ax.set_ylabel("")
#ax.set_yticklabels([])


ax = fig.add_subplot(1, 3, 2)
momentsml.plot.scatter.scatter(ax, cat, tru_s2, pre_s2_wbias, showidline=True, idlinekwargs=idlinekwargs, yisres=True)
addmetrics(ax, tru_s2, pre_s2_wbias)

#ax.set_ylabel("")
#ax.set_yticklabels([])


if mode != 3:
	ax = fig.add_subplot(1, 3, 3)
	cb = momentsml.plot.scatter.scatter(ax, cat, tru_mag, tru_rad, snr, cmap="plasma_r", rasterized = True, norm=matplotlib.colors.LogNorm(vmin=3.0, vmax=150.0))
	ticks = [5, 10, 20, 40, 80]
	ticklabels = [str(tick) for tick in ticks]
	cb.set_ticks(ticks)
	cb.ax.set_yticklabels(ticklabels)


fig.text(.005, .94, title, fontdict={"fontsize":12})

momentsml.plot.figures.savefig(os.path.join(config.valdir, valname.replace(".", "p") + "_selbiaseffect"), fig, fancy=True, pdf_transparence=True, nocrop=True)
plt.show()






import numpy as np
import tenbilac

import logging
logging.basicConfig(level=logging.INFO)


# 2 inputs and 3 neurons

layer = tenbilac.layer.Layer(ni=2, nn=3, actfct=tenbilac.act.iden, mode="mult")

#layer.addnoise()
#print layer.report()

#layer.zero()
#print layer.report()

# First neuron: simple product
layer.weights[0, 0] = 1
layer.weights[0, 1] = 1

# Second neuron: product of almost-squares
layer.weights[1, 0] = 2.1
layer.weights[1, 1] = 2.1

# Third one is left to play 
layer.weights[2, 0] = 1.0
layer.weights[2, 1] = -1.0

layer.biases[2] = 42

print layer.report()

print "1D case, just a single case or realization"

inp = np.array([-1.0, 2.0])

outp = layer.run(inp)

print "input = ", inp

print "ouput = ", outp


print "2D input: (feature=2, case=4identical)"

inp = np.array([[-1.0, -1.0, -1.0, -1.0], [2.0, 2.0, 2.0, 2.0]])

outp = layer.run(inp)

print "input = ", inp

print "ouput = ", outp, "shape", outp.shape



print "3D input: (realization=5, feature=2, case=4)"

inp = np.array([
[[-1.0, -1.0, -1.0, -1.0], [2.0, 2.0, 2.0, 2.0]],
[[-1.0, -1.0, -1.0, -1.0], [2.0, 2.0, 2.0, 2.0]],
[[-1.0, -1.0, -1.0, -1.0], [2.0, 2.0, 2.0, 2.0]],
[[-1.0, -1.0, -1.0, -1.0], [2.0, 2.0, 2.0, 2.0]],
[[-1.0, -1.0, -1.0, -1.0], [2.0, 2.0, 2.0, 2.0]]
])

outp = layer.run(inp)

print "input = ", inp

print "ouput = ", outp, "shape",  outp.shape




#import matplotlib
#matplotlib.use("AGG")

import matplotlib.pyplot as plt

import momentsml
import momentsml.tools
import momentsmlgreat3
import astropy


import config
import numpy as np
import os


import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)


subfields = []
psfsigmas = []
nfails = []


for subfield in config.great3.subfields:
	
	predcatpath = config.great3.subpath(subfield, "pred", "predcat_{}.pkl".format(config.predcode))
	cat = momentsml.tools.io.readpickle(predcatpath)
	
	#print momentsml.tools.table.info(cat)
	#n = float(len(cat))
	#print subfield, np.sum(cat["pre_s1"].mask), np.sum(cat["adamom_g1"].mask)
	
	subfields.append(subfield)
	nfails.append(np.sum(cat["pre_s1"].mask))
	psfsigmas.append(np.mean(cat["psf_adamom_sigma"]))
	
	
plt.scatter(psfsigmas, nfails)
plt.show()



"""
Same as run_32, but uses weights, if available
"""

import matplotlib
matplotlib.use("AGG")

import momentsml.tools
import momentsml.learn
import momentsml

import config
import numpy as np
import os

import plot_2_val

import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)






for subfield in config.great3.subfields:

	logger.info("Working on subfield {}".format(subfield))

	
	catpath = config.great3.subpath(subfield, "simmeas", config.datasets["valid-shear"], "groupmeascat.pkl")
	cat = momentsml.tools.io.readpickle(catpath)
		
	traindir = config.great3.subpath(subfield, "ml", config.datasets["train-shear"])	
	predcat = momentsml.learn.tenbilacrun.predict(cat, config.shearconflist, traindir)

	if len(config.weightconflist) > 0:
		
		weighttraindir = config.great3.subpath(subfield, "ml", config.datasets["train-weight"])
		predcat = momentsml.learn.tenbilacrun.predict(predcat, config.weightconflist , weighttraindir)

	
	shearconfnames = momentsml.learn.tenbilacrun.confnames(config.shearconflist)
	for (confname, conf) in zip(shearconfnames, config.shearconflist):

		predcatcopy = predcat.copy() # As plots modifies it.
		
		if "s2" in conf[0] or "g2" in conf[0]:
			component = 2
		else:
			component = 1
		
		valname = "pred_{}_with_{}_{}_weights".format(config.datasets["valid-shear"], config.datasets["train-shear"], confname)
		figpredcatpath = config.great3.subpath(subfield, "val", valname + ".png")
		#figpath = config.great3.subpath(subfield, "val", "33_valw_{}.png".format(component))
		
		plot_2_val.plot(predcatcopy, component, mode="s", withweights=True, filepath=figpredcatpath)
		
	



import tenbilac


import logging
logging.basicConfig(level=logging.INFO)



net = tenbilac.net.Net(ni=2, nhs=[3, 3], onlyid=True)
net.setidentity()

params = net.get_params_ref()

print params

net.addnoise()

print params

net.setidentity()

print params

print net.report()




import momentsml.meas


def default(catalog, stampsize):
	"""
	Default measfct, runs on "img".
	"""	
	
	# HSM adamom
	catalog = momentsml.meas.galsim_adamom.measfct(catalog, stampsize=stampsize, variant="wider")
	catalog = momentsml.meas.adamom_calc.measfct(catalog)
	
	# And skystats
	catalog = momentsml.meas.skystats.measfct(catalog, stampsize=stampsize)
	
	# And snr
	catalog = momentsml.meas.snr.measfct(catalog, gain=1.0e12) # Gain set to give sky-limited SNR
	
	
	return catalog
	
	


default_groupcols = [
'adamom_flag',
'adamom_flux',
'adamom_x',
'adamom_y',
'adamom_g1',
'adamom_g2',
'adamom_sigma',
'adamom_rho4',
'adamom_logflux',
'adamom_g',
'adamom_theta',
'skystd',
'skymad',
'skymean',
'skymed',
'skystampsum',
'skyflag',
'snr'
]

default_removecols = []



import matplotlib
matplotlib.use("AGG")

import momentsml.tools
import momentsml.learn
import momentsml

import config
import numpy as np
import os


import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)

"""
# Replaced by config
# Build a name to save the overall validation: 
#sconfname = os.path.splitext(os.path.basename(config.shearconflist[0][1]))[0] # extracts e.g. "sum55"
#wconfname = os.path.splitext(os.path.basename(config.weightconflist[0][1]))[0] # extracts e.g. "sum55w"
#valname = "{}_and_{}_with_{}_{}_on_{}".format(config.datasets["ts"], config.datasets["tw"], sconfname, wconfname, config.datasets["vo"])
"""


# Load validation catalog

catpath = os.path.join(config.simmeasdir, config.datasets["vo"], "groupmeascat.pkl")
cat = momentsml.tools.io.readpickle(catpath)

# Predict point estimates:

traindir = os.path.join(config.traindir, config.datasets["tp"])
predcat = momentsml.learn.tenbilacrun.predict(cat, config.shearconflist, traindir)


# Predict weights:
wtraindir = os.path.join(config.traindir, config.datasets["tw"] + "_with_" + config.datasets["tp"] + "_" + config.sconfname)
predcat = momentsml.learn.tenbilacrun.predict(predcat, config.weightconflist, wtraindir)


# Save the prediction:

predcatpath = os.path.join(config.valdir, config.wvalname+".pkl")
momentsml.tools.io.writepickle(predcat, predcatpath)





import galsim
import numpy as np
import os
import astropy
import os
import config
import matplotlib.pyplot as plt


"""
If psfoversampling is 10,
the resulting PSF has a FWHM of about 10 pixels (measured with skycat or fv) -> 1 VIS pixel -> 0.1 arcsec
This is smaller than what can be found in Cropper 2016, but the pixel convolution is missing here!
"""



filterband_min = 550. 
filterband_max = 900.


spectrum = astropy.table.Table.read(config.spectrumpath)

# We compute some new columns:

spectrum["lam"] = spectrum["WAVELENGTH"] * 0.1 # in nm

# To weight the PSFs, we want something prop to number of photons
# The column "FLUX" is in ergs/s/cm/A
# The energy of a photon is in 1/lam
# -> We divide the flux by energy per photon

spectrum["nphot"] = spectrum["FLUX"] * spectrum["WAVELENGTH"]
spectrum["nphot"] = spectrum["nphot"] / np.max(spectrum["nphot"])

# We define the indices of the table over which we sum
mask = np.logical_and(spectrum["lam"] >= filterband_min, spectrum["lam"] <= filterband_max)
indices = np.where(mask)[0]

indices = indices[::50] # Reduce the number of indices
print "Number of indices to sum over: ", len(indices)
 

plt.plot(spectrum["lam"], spectrum["nphot"])
plt.plot(spectrum["lam"][indices], spectrum["nphot"][indices], "o")
plt.show()

images = []

for i, index in enumerate(indices):
	
	lam = spectrum["lam"][index] 
	nphot = spectrum["nphot"][index] 
	
	print index, lam, nphot
	
	psf = galsim.OpticalPSF(lam=lam,
		diam=1.2,
		obscuration=0.29,
		nstruts=6,		
		flux = 1.0
		)
		
	image = galsim.ImageF(config.psfstampsize, config.psfstampsize)
	psf.drawImage(image=image, scale=0.1/config.psfoversampling)
	
	image *= nphot
	
	images.append(image)

image = sum(images)

totflux = np.sum(image.array)
image /= totflux

print "Output image flux:", np.sum(image.array)


image.write(os.path.join(config.workdir, "psf.fits"))




"""
Classic simobscompa plot
"""
import matplotlib
matplotlib.use("AGG")

import momentsml
import momentsmlgreat3
import momentsml.plot
from momentsml.tools.feature import Feature
import matplotlib.pyplot as plt
import numpy as np

import config

import os
import logging
logging.basicConfig(format=config.loggerformat, level=logging.INFO)
logger = logging.getLogger(__name__)



def main():

	
	for subfield in config.great3.subfields:
		
		spname = config.datasets["simobscompa"]
		plotpath = config.great3.subpath(subfield, "simmeas", "{}.png".format(spname))
		#plotpath = None
		if config.great3.skipdone and os.path.exists(plotpath):
			logger.info("Subfield {} is already done".format(subfield))
			continue
		
		measdir = config.great3.subpath(subfield, "simmeas")

		simcat = momentsml.tools.io.readpickle(os.path.join(measdir, spname, "groupmeascat.pkl"))
		#print momentsml.tools.table.info(simcat)


		obscat = momentsml.tools.io.readpickle(config.great3.subpath(subfield, "obs", "img_meascat.pkl"))
		#print momentsml.tools.table.info(obscat)
		#obscat = obscat[:1000]
			
		plot(simcat, obscat, filepath=plotpath)
		logger.info("Plotted to {}".format(plotpath))




def plot(simcat, obscat, filepath=None):

	# Some computations
	simcat["aperphot_sbr1"] = simcat["aperphot_sb2"] / simcat["aperphot_sb5"]
	obscat["aperphot_sbr1"] = obscat["aperphot_sb2"] / obscat["aperphot_sb5"]

	simcat["aperphot_sbr2"] = simcat["aperphot_sb3"] / simcat["aperphot_sb8"]
	obscat["aperphot_sbr2"] = obscat["aperphot_sb3"] / obscat["aperphot_sb8"]

	simcat["aperphot_log_sb2"] = np.log10(simcat["aperphot_sb2"])
	simcat["aperphot_log_sb5"] = np.log10(simcat["aperphot_sb5"])
	obscat["aperphot_log_sb2"] = np.log10(obscat["aperphot_sb2"])
	obscat["aperphot_log_sb5"] = np.log10(obscat["aperphot_sb5"])


	simcat["adamom_log_flux"] = np.log10(simcat["adamom_flux"])
	obscat["adamom_log_flux"] = np.log10(obscat["adamom_flux"])

	
	rea = None

	adamom_flux = Feature("adamom_flux", rea=rea)
	adamom_sigma = Feature("adamom_sigma", 0, 10, rea=rea)
	adamom_rho4 = Feature("adamom_rho4", 1.3, 3.0, rea=rea)
	adamom_g1 = Feature("adamom_g1", -0.7, 0.7, rea=rea)
	adamom_g2 = Feature("adamom_g2", -0.7, 0.7, rea=rea)
	adamom_log_flux = Feature("adamom_log_flux", rea=rea)
	aperphot_sbr1 = Feature("aperphot_sbr1", rea=rea)
	aperphot_sbr2 = Feature("aperphot_sbr2", rea=rea)
	snr = Feature("snr", -3, 30, rea=rea)
	aperphot_sb2 = Feature("aperphot_sb2", 0, 5, rea=rea)
	aperphot_sb3 = Feature("aperphot_sb3", 0, 5, rea=rea)
	aperphot_sb5 = Feature("aperphot_sb5", 0, 5, rea=rea)
	aperphot_sb8 = Feature("aperphot_sb8", 0, 5, rea=rea)
	aperphot_log_sb5 = Feature("aperphot_log_sb5", rea=rea)
	aperphot_log_sb2 = Feature("aperphot_log_sb2", rea=rea)
	skymad = Feature("skymad", rea=rea)
	skystd = Feature("skystd", rea=rea)
	skymed = Feature("skymed", rea=rea)
	skymean = Feature("skymean", rea=rea)

	#psf_adamom_g1 = Feature("psf_adamom_g1", -0.06, 0.06, rea=rea)
	#psf_adamom_g2 = Feature("psf_adamom_g2", -0.06, 0.06, rea=rea)
	#psf_adamom_sigma = Feature("psf_adamom_sigma", rea=rea)





	fig = plt.figure(figsize=(24, 14))
	#fig = plt.figure(figsize=(8, 8))

	ax = fig.add_subplot(3, 5, 1)

	momentsml.plot.contour.simobs(ax, simcat, obscat, adamom_g1, adamom_g2, plotpoints=False, nlines=2)
	#momentsml.plot.hist.hist(ax, simcat, snr, color="red", label="Training", normed=True)
	#momentsml.plot.hist.hist(ax, obscat, snr, color="blue", label="GREAT3", normed=True)

	ax = fig.add_subplot(3, 5, 2)
	momentsml.plot.scatter.simobs(ax, simcat, obscat, adamom_g1, adamom_g2)
	
	ax = fig.add_subplot(3, 5, 3)
	momentsml.plot.scatter.simobs(ax, simcat, obscat, snr, adamom_sigma, legend=True)

	ax = fig.add_subplot(3, 5, 4)
	momentsml.plot.scatter.simobs(ax, simcat, obscat, adamom_flux, adamom_sigma)

	ax = fig.add_subplot(3, 5, 5)
	momentsml.plot.scatter.simobs(ax, simcat, obscat, Feature("adamom_flux", 0, 80, rea=rea), Feature("adamom_sigma", 0.5, 3.5, rea=rea))
	

	ax = fig.add_subplot(3, 5, 6)
	momentsml.plot.scatter.simobs(ax, simcat, obscat, aperphot_sb2, aperphot_sb3)

	ax = fig.add_subplot(3, 5, 7)
	momentsml.plot.scatter.simobs(ax, simcat, obscat, aperphot_sb5, aperphot_sb8)

	ax = fig.add_subplot(3, 5, 8)
	momentsml.plot.scatter.simobs(ax, simcat, obscat, aperphot_sbr1, aperphot_sbr2)

	ax = fig.add_subplot(3, 5, 9)
	momentsml.plot.scatter.simobs(ax, simcat, obscat, adamom_log_flux, adamom_sigma)




	ax = fig.add_subplot(3, 5, 11)
	momentsml.plot.contour.simobs(ax, simcat, obscat, skymad, skymean, plotpoints=False)

	ax = fig.add_subplot(3, 5, 12)
	momentsml.plot.scatter.simobs(ax, simcat, obscat, aperphot_log_sb2, aperphot_log_sb5)

	ax = fig.add_subplot(3, 5, 13)
	momentsml.plot.scatter.simobs(ax, simcat, obscat, adamom_rho4, adamom_sigma)

	ax = fig.add_subplot(3, 5, 14)
	momentsml.plot.scatter.simobs(ax, simcat, obscat, adamom_log_flux, adamom_rho4)
	#ax.set_xscale("log", nonposx='clip')

	
	#ax = fig.add_subplot(3, 4, 12)
	#momentsml.plot.scatter.simobs(ax, simcat, obscat, psf_adamom_g1, psf_adamom_g2)

	#ax = fig.add_subplot(3, 4, 2)
	#momentsml.plot.scatter.scatter(ax, cat, Feature("tru_s1"), Feature("tru_s2"))

	
	
	plt.tight_layout()

	if filepath:
		plt.savefig(filepath)
	else:
		plt.show()
	plt.close(fig) # Helps releasing memory when calling in large loops.


if __name__ == "__main__":
    main()
	



"""
Functions to group and summarize measurements accross multiple realizations of simulations.

"""
import os
import sys

import astropy
import numpy as np
import datetime

from .. import tools
import utils

import logging
logger = logging.getLogger(__name__)

def onsims(measdir, simparams, task="groupstats", **kwargs):
	"""
	Top-level function to group measurements as obtained from :func:`momentsml.meas.run.onsims`.
	
	This function first explores the files in your measdir using :func:`simmeasdict`,
	then uses :func:`groupstats` or :func:`group` to "hstack" the different realizations catalog by catalog (and maybe compute some statistics),
	and finally it "vstacks" all your catalogs to return a single output table.

	:param measdir: See :func:`momentsml.meas.run.onsims`
	:param simparams: idem
	:param task: either "groupstats" or "group", depending on what you want to do.
	:param kwargs: any further keyword arguments are passed to :func:`momentsml.tools.table.groupstats` or :func:`momentsml.tools.table.group`
	
	To learn about the output, see the latter.
	
	"""
	
	logger.info("Harvesting measurements of the simulations '%s' from %s" % (simparams, measdir))
	# We use the above function to find what we have:
	catdict = utils.simmeasdict(measdir, simparams)
	
	# We iterate over one simulated catalog after the other.
	outputcats = []
	for (catname, meascatfilepaths) in catdict.items():
		
		# So this are the measurements for a single catalog:
		logger.info("Reading all measurements for catalog '%s' (%i realizations)..." % (catname, len(meascatfilepaths)))
		meascats = [tools.io.readpickle(os.path.join(measdir, simparams.name, meascatfilepath)) for meascatfilepath in meascatfilepaths]
		
		# To avoid meta conflicts, we remove the meta from all but the first realization:
		for meascat in meascats[1:]:
			meascat.meta = {}
		# We also remove meta["img"] from the first realization, as it does no longer apply after the groupstats.
		meascats[0].meta.pop("img")
		
		if task == "groupstats":
			# And we call groupstats
			logger.info("Grouping columns and computing averages for catalog '%s'..." % (catname))
			grouped = tools.table.groupstats(meascats, **kwargs)
			outputcats.append(grouped)
		
		elif task == "group":
			logger.info("Grouping columns for catalog '%s'..." % (catname))
			grouped = tools.table.group(meascats, **kwargs)
			outputcats.append(grouped)
	
		else:
			raise RuntimeError("Unknown task!")	
	
	# And finally, we stack all these catalogs "vertically", to get a single one.
	# This is only to be done if we have several catalogs. In fact, vstack crashes if only 
	# one catalog is present.
	if len(outputcats) > 1:
	
		# Before vstacking, we check some meta for compatibility
		nmetaname = "n" + task # So this is either ngroup or ngroupstats
		assert nmetaname in outputcats[0].meta
			
		# All catalogs should have the same number of realizations
			
		ngroupstats = outputcats[0].meta[nmetaname]
		for outputcat in outputcats:
			if outputcat.meta[nmetaname] != ngroupstats:
				raise RuntimeError("Catalogs with different numbers of realizations should not be merged, clean your measdir accordingly!")
		
		# Now that checks are done, we have to remove some of the meta, to avoid conflicts (and keep thinks logical)
		for outputcat in outputcats[1:]: # On all but the first one, we remove everything.
			outputcat.meta = {}
		outputcats[0].meta.pop("catname", None) # after the merge, a single catname has no meaning anymore
		outputcats[0].meta.pop("psf", None) # There is no single "psf" anymore.
		outputcats[0].meta.pop("imgreas", None) # Idem
		
		logger.info("Concatenating catalogs...")
		outputcat = astropy.table.vstack(outputcats, join_type="exact", metadata_conflicts="error")
	
	else:
		logger.debug("Only one catalog, no need to call vstack")
		outputcat = outputcats[0]
	
	logger.info("Done with collecting results for %i simulated galaxies" % (len(outputcat)))
	
	return outputcat





"""
This files holds a user-defined measfct.

It's a very good idea to keep this function in a separate module and not in your
script, so that it can be imported by the multiprocessing pool workers that will call,
it without causing any trouble!
"""

import momentsml.meas
import momentsml.meas.sewfunc
import momentsml.meas.aperphot
import config


def psf(catalog, branch=None):
	"""
	This is to measure PSF shapes when running on the PSF catalog, and therefore the PSF image stamps
	are to be referenced in "img", not in "psf"...
	"""	
	
#	sewpy_config = {"DETECT_MINAREA":6, "DETECT_THRESH":2, "ANALYSIS_THRESH":2,
#		"PHOT_FLUXFRAC":"0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9",
#		"ASSOC_RADIUS":5, "ASSOC_TYPE":"NEAREST"}
#	
#	sewpy_params = ["VECTOR_ASSOC(3)", "XWIN_IMAGE", "YWIN_IMAGE", "AWIN_IMAGE", "BWIN_IMAGE", "THETAWIN_IMAGE",
#		"FLUX_WIN", "FLUXERR_WIN", "NITER_WIN", "FLAGS_WIN", "FLUX_AUTO", "FLUXERR_AUTO",
#		"FWHM_IMAGE", "KRON_RADIUS", "FLUX_RADIUS(7)", "BACKGROUND", "FLAGS"]
#	
#	catalog = momentsml.meas.sewfunc.measfct(catalog, runon="img", config=sewpy_config,
#		params=sewpy_params, sexpath=config.sexpath,
#		prefix="psf_sewpy_")
	
	# We run galsim_adamom :
	catalog = momentsml.meas.galsim_adamom.measfct(catalog, stampsize=branch.stampsize(), prefix="psf_adamom_", variant="wider")
	
	# We run skystats:
	catalog = momentsml.meas.skystats.measfct(catalog, stampsize=branch.stampsize(), prefix="psf_")
			
	return catalog




def gal(catalog, branch=None):
	"""
	The measfct for galaxies.
	
	"""	
	
#	sewpy_config = {"DETECT_MINAREA":6, "DETECT_THRESH":2, "ANALYSIS_THRESH":2,
#		"PHOT_FLUXFRAC":"0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9",
#		"ASSOC_RADIUS":5, "ASSOC_TYPE":"NEAREST"}
#	
#	sewpy_params = ["VECTOR_ASSOC(3)", "XWIN_IMAGE", "YWIN_IMAGE", "AWIN_IMAGE", "BWIN_IMAGE", "THETAWIN_IMAGE",
#		"FLUX_WIN", "FLUXERR_WIN", "NITER_WIN", "FLAGS_WIN", "FLUX_AUTO", "FLUXERR_AUTO",
#		"FWHM_IMAGE", "KRON_RADIUS", "FLUX_RADIUS(7)", "BACKGROUND", "FLAGS"]
#	
#	catalog = momentsml.meas.sewfunc.measfct(catalog, runon="img", config=sewpy_config,
#		params=sewpy_params, sexpath=config.sexpath)
	
	# We run galsim_adamom :
	catalog = momentsml.meas.galsim_adamom.measfct(catalog, stampsize=branch.stampsize(), variant="wider")
	
	# We run skystats:
	catalog = momentsml.meas.skystats.measfct(catalog, stampsize=branch.stampsize())
	
	# aperphot:
	catalog = momentsml.meas.aperphot.measfct(catalog, radii=(2, 3, 5, 8))
	
	# and snr
	catalog = momentsml.meas.snr.measfct(catalog, gain=1.0)
		
	
	return catalog



groupcols = [
 'adamom_flag',
 'adamom_flux',
 'adamom_x',
 'adamom_y',
 'adamom_g1',
 'adamom_g2',
 'adamom_sigma',
 'adamom_rho4',
 'skyflag',
 'skystd',
 'skymad',
 'skymean',
 'skymed',
 'skystampsum',
 'aperphot_sb2',
 'aperphot_sb3',
 'aperphot_sb5',
 'aperphot_sb8',
 'snr'
]

removecols = [
		
]



"""
# Some previously used values:

'sewpy_XWIN_IMAGE',
'sewpy_YWIN_IMAGE',
'sewpy_AWIN_IMAGE',
'sewpy_BWIN_IMAGE',
'sewpy_THETAWIN_IMAGE',
'sewpy_FLUX_WIN',
'sewpy_FLUXERR_WIN',
'sewpy_NITER_WIN',
'sewpy_FLAGS_WIN',
'sewpy_FLUX_AUTO',
'sewpy_FLUXERR_AUTO',
'sewpy_FWHM_IMAGE',
'sewpy_KRON_RADIUS',
'sewpy_FLUX_RADIUS',
'sewpy_FLUX_RADIUS_1',
'sewpy_FLUX_RADIUS_2',
'sewpy_FLUX_RADIUS_3',
'sewpy_FLUX_RADIUS_4',
'sewpy_FLUX_RADIUS_5',
'sewpy_FLUX_RADIUS_6',
'sewpy_BACKGROUND',
'sewpy_FLAGS',
'sewpy_assoc_flag',


['id',
 'ix',
 'iy',
 'snc_type',
 'tru_flux',
 'tru_g1',
 'tru_g2',
 'tru_gain',
 'tru_mu',
 'tru_pixel',
 'tru_rad',
 'tru_read_noise',
 'tru_s1',
 'tru_s2',
 'tru_sersicn',
 'tru_sky_level',
 'tru_type',
 'x',
 'y',
 'psfx',
 'psfy',
 'psf_adamom_flag',
 'psf_adamom_flux',
 'psf_adamom_x',
 'psf_adamom_y',
 'psf_adamom_g1',
 'psf_adamom_g2',
 'psf_adamom_sigma',
 'psf_adamom_rho4',
 'psf_skyflag',
 'psf_skystd',
 'psf_skymad',
 'psf_skymean',
 'psf_skymed',
 'psf_skystampsum',
 'adamom_flag',
 'adamom_flux',
 'adamom_x',
 'adamom_y',
 'adamom_g1',
 'adamom_g2',
 'adamom_sigma',
 'adamom_rho4',
 'skyflag',
 'skystd',
 'skymad',
 'skymean',
 'skymed',
 'skystampsum',
 'aperphot_sb2',
 'aperphot_sb3',
 'aperphot_sb5',
 'aperphot_sb8',
 'snr']

"""





"""
This module provides several specific functions to save figures nicely. `fancy` means latex 
interpreter and generates *.pdf files
"""

import numpy as np
import matplotlib

def get_colors():
	return ['k', 'orange', 'skyblue', 'lightgrey', "royalblue", 'r', "#a65628", "#f781bf", "#4daf4a"]
	#return ['k', 'orange', 'skyblue', 'seagreen', "#984ea3", 'r', "#a65628", "#f781bf", "#4daf4a"]

def savefig(fname,fig,fancy=False,pdf_transparence=True, nocrop=False):
	import os
	import subprocess

	directory=os.path.dirname(os.path.abspath(fname))
	if not os.path.exists(directory):
		os.makedirs(directory)

	fig.savefig(fname+'.png',dpi=300)

	if fancy: 
		fig.savefig(fname+'.pdf',transparent=pdf_transparence)
		#fig.savefig(fname+'.eps',transparent=True)
		#os.system("epstopdf "+fname+".eps")
		if nocrop is False:
			command = 'pdfcrop %s.pdf' % fname
			subprocess.check_output(command, shell=True)
			os.system('mv '+fname+'-crop.pdf '+fname+'.pdf')
	

def set_fancy(txtsize=16):
	from matplotlib import rc
	#rc('font',**{'family':'serif','serif':['Palatino'],'size':16})
	rc('font',**{'size':txtsize})
	#rc('font', **{'family': 'serif', 'serif': ['Computer Modern']})
	rc('font', **{'family':'sans-serif','sans-serif':['Arial']})
	rc('text', usetex=True)


def cmap_map(function,cmap):
	""" Applies function (which should operate on vectors of shape 3:
	[r, g, b], on colormap cmap. This routine will break any discontinuous	 points in a colormap.
	"""
	cdict = cmap._segmentdata
	step_dict = {}
	# Firt get the list of points where the segments start or end
	for key in ('red','green','blue'):		 step_dict[key] = map(lambda x: x[0], cdict[key])
	step_list = sum(step_dict.values(), [])
	step_list = np.array(list(set(step_list)))
	# Then compute the LUT, and apply the function to the LUT
	reduced_cmap = lambda step : np.array(cmap(step)[0:3])
	old_LUT = np.array(map( reduced_cmap, step_list))
	new_LUT = np.array(map( function, old_LUT))
	# Now try to make a minimal segment definition of the new LUT
	cdict = {}
	for i,key in enumerate(('red','green','blue')):
		this_cdict = {}
		for j,step in enumerate(step_list):
			if step in step_dict[key]:
				this_cdict[step] = new_LUT[j,i]
			elif new_LUT[j,i]!=old_LUT[j,i]:
				this_cdict[step] = new_LUT[j,i]
		colorvector=  map(lambda x: x + (x[1], ), this_cdict.items())
		colorvector.sort()
		cdict[key] = colorvector

	return matplotlib.colors.LinearSegmentedColormap('colormap',cdict,1024)




"""
This generates the "submission file" (for constant shear branches), bypassing the GREAT3 "presubmission script".
"""


import momentsml
import momentsml.tools

import astropy

import config
import numpy as np
import os


import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)

catpath = config.great3.path("summary_{}.pkl".format(config.predcode))
cat = momentsml.tools.io.readpickle(catpath)

subcat = cat["subfield","pre_s1w","pre_s2w"]

outcatpath = config.great3.path("submission_{}.txt".format(config.predcode))
logger.info("Writing {}...".format(outcatpath))
subcat.write(outcatpath, format="ascii.commented_header")




"""
Configuration for the GREAT3 scripts.
"""

import momentsmlgreat3 as mg3

loggerformat='\033[1;31m%(levelname)s\033[1;0m: %(name)s(%(funcName)s): \033[1;21m%(message)s\033[1;0m'


great3 = mg3.great3.GREAT3Run(
	experiment = "control",
	obstype = "space",
	sheartype = "constant",
	datadir = "/vol/fohlen11/fohlen11_1/mtewes/GREAT3",
	truthdir = "/vol/fohlen11/fohlen11_1/mtewes/GREAT3/truth", # Only needed for final analysis plots
	workdir = "/vol/fohlen11/fohlen11_1/mtewes/2017_MomentsML_GREAT3/csc_v2",
	g3publicdir = "/users/mtewes/code/great3-public",
	
	#subfields = [94],
	#subfields = [45],	# CSC best PSF subfields: [45, 75, 136],  worst : [143, 49, 150], 94 could not be measured!
	
	#subfields = [1045],
	#subfields = [0],
	
	subfields = range(0, 200),
	#subfields = range(0, 50),
	#subfields = range(50, 100),
	#subfields = range(100, 150),
	#subfields = range(150, 200),
	
	
	#subfields = range(84, 100),
	
	ncpu = 4,
	skipdone = False
	)



### Script configuration ###

datasets = {
	"train-shear":"ts-ell-nn-train-rea10",
	#"train-shear":"ts-ell-n-train-rea100",
	
		
	"valid-shear":"vs-shear-n-G3-snc1000",
	"train-weight":"tw-200c-1000r",
	#"train-weight":"tw-100c-10000r",
	
	"valid-overall":"vo-200c-8000r",
	#"simobscompa":"simobscompa-G3",
	"simobscompa":"simobscompa-train",
	
	"mimic-great3":"sersicG3subfield"
	#"mimic-great3":"sersicG3subfield_nosnc"
	
}


shearconflist = [
	("mlconfig/ada4g1.cfg", "mlconfig/sum55mass.cfg"), # Comment a line to run on only one component
	("mlconfig/ada4g2.cfg", "mlconfig/sum55mass.cfg")
	#("mlconfig/ada4g1.cfg", "mlconfig/sum55mab.cfg"), # Comment a line to run on only one component
	#("mlconfig/ada4g2.cfg", "mlconfig/sum55mab.cfg")


]

weightconflist = [
	# Used for mass:
	#("mlconfig/ada4s1w.cfg", "mlconfig/sum33wmassshort.cfg"),  ## predcode 1
	#("mlconfig/ada4s2w.cfg", "mlconfig/sum33wmassshort.cfg"),
	
	#("mlconfig/ada4s1w.cfg", "mlconfig/sum33wmab.cfg"),
	#("mlconfig/ada4s2w.cfg", "mlconfig/sum33wmab.cfg"),
	
	("mlconfig/ada5s1w.cfg", "mlconfig/sum55wmass500.cfg"),  ## predcode 2
	("mlconfig/ada5s2w.cfg", "mlconfig/sum55wmass500.cfg"),

	
]

#weightconflist = []

trainmode = "g" # When training for ellipticity, switch this to "g". It's for the plots.

predcode = "2" # String used in file names of GREAT3 predictions. Allows you to keep several predictions side by side.



import momentsml
import momentsml.plot

import os
import config
import numpy as np

from momentsml.tools.feature import Feature

import matplotlib
import matplotlib.pyplot as plt



cat = momentsml.tools.io.readpickle(os.path.join(config.simmeasdir, config.datasets["si"], "groupmeascat.pkl"))

print momentsml.tools.table.info(cat)

fig = plt.figure(figsize=(20, 13))


ax = fig.add_subplot(2, 3, 1)
momentsml.plot.scatter.scatter(ax, cat, Feature("tru_psf_x"),  Feature("tru_psf_y"),  Feature("tru_psf_g1"))
ax = fig.add_subplot(2, 3, 2)
momentsml.plot.scatter.scatter(ax, cat, Feature("tru_psf_x"),  Feature("tru_psf_y"),  Feature("tru_psf_g2"))
ax = fig.add_subplot(2, 3, 3)
momentsml.plot.scatter.scatter(ax, cat, Feature("tru_psf_x"),  Feature("tru_psf_y"),  Feature("tru_psf_sigma"))



ax = fig.add_subplot(2, 3, 4)
momentsml.plot.scatter.scatter(ax, cat, Feature("tru_psf_x"), Feature("tru_psf_y"), Feature("adamom_g1"))
ax = fig.add_subplot(2, 3, 5)
momentsml.plot.scatter.scatter(ax, cat, Feature("tru_psf_x"), Feature("tru_psf_y"), Feature("adamom_g2"))
ax = fig.add_subplot(2, 3, 6)
momentsml.plot.scatter.scatter(ax, cat, Feature("tru_psf_x"), Feature("tru_psf_y"), Feature("adamom_sigma"))


plt.tight_layout()
plt.show()
plt.close(fig) # Helps releasing memory when calling in large loops.





import matplotlib
matplotlib.use("AGG")

import momentsml.tools
import momentsml.learn
import momentsml

import config
import numpy as np
import os


import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)




catpath = os.path.join(config.simmeasdir, config.datasets["tw"], "groupmeascat.pkl")
cat = momentsml.tools.io.readpickle(catpath)

#print momentsml.tools.table.info(cat)
#exit()

wtraindir = os.path.join(config.traindir, config.datasets["tw"] + "_with_" + config.datasets["tp"] + "_" + config.sconfname)
if not os.path.exists(wtraindir):
	os.makedirs(wtraindir)

predcatpath = os.path.join(wtraindir, "groupmeascat_predforw.pkl")


traindir = os.path.join(config.traindir, config.datasets["tp"])

	
predcat = momentsml.learn.tenbilacrun.predict(cat, config.shearconflist, traindir)

momentsml.tools.io.writepickle(predcat, predcatpath)
	
	



"""
Demo corresponding to Appendix 1 of our paper (https://arxiv.org/abs/1807.02120)
"""

import numpy as np
import tenbilac

import logging
logging.basicConfig(level=logging.INFO)


# We define the noise distribution:
def noise(n):
	return np.random.randn(n)
	#return 0.1*(np.exp(0.4*np.random.randn(n)) - 0.7)

# And the function behind the data:
def f(x):
	return np.sqrt(1.0 + x**2)


# Amount of data that will be used for training (to make it faster, reduce number of cases...)
train_ncase = 1000 # Number of cases (i.e., number of parameters "theta")
train_nrea = 1000 # How many realizations per case (i.e., "observations" per parameter).
noise_scale = 0.1


# Now we  generate the data, in form of numpy arrays (we don't use masked arrays here, but we could, if needed).
# Important general note for using tenbilac:
# - the *input* data (called here "obs", or "d") has to be 3D. 
#   The indices are: (realization, node of the input layer, case)
# - the *target* data (called here "params", or "theta") has to be 2D
#   The indices are: (node of the output layer, case)
# For the present example, we will use a network with only one input node, and only one output node.


# The explanatory variables:
train_params = np.linspace(0.25, 2.0, train_ncase).reshape((1, train_ncase))
# And the corresponding observations of the dependt variable:
train_obs = np.array([f(train_params) + noise_scale*noise(train_ncase).reshape((1, train_ncase)) for rea in range(train_nrea)])


# To study bias afterwards, we make a "validation" set.
val_ncase = 1000
val_nrea = 1000
val_params = np.linspace(0.25, 2.0, val_ncase).reshape((1, val_ncase))
val_obs = np.array([f(val_params) + noise_scale*noise(val_ncase).reshape((1, val_ncase)) for rea in range(val_nrea)])


# To plot the inverse regression as continuous lines, it's good to have samples uniform in the dependent variable:
test_obs = np.linspace(0.5, 3, 100).reshape((1, 100))


# Now we norm all these. We build the Normers by using the larger validation data (but training would work as well...):
obs_normer = tenbilac.data.Normer(val_obs)
params_normer = tenbilac.data.Normer(val_params)

# And apply them:
normed_train_params = params_normer(train_params)
normed_train_obs = obs_normer(train_obs)
normed_val_params = params_normer(val_params)
normed_val_obs = obs_normer(val_obs)
normed_test_obs = obs_normer(test_obs)

# And save a pkl file


pkldata = {
	"obs_normer":obs_normer,
	"params_normer":params_normer,
	"train_params":train_params,
	"train_obs":train_obs,
	"normed_train_params":normed_train_params,
	"normed_train_obs":normed_train_obs,
	"val_params":val_params,
	"val_obs":val_obs,
	"normed_val_params":normed_val_params,
	"normed_val_obs":normed_val_obs,
	"test_obs":test_obs,
	"normed_test_obs":normed_test_obs,
}

tenbilac.utils.writepickle(pkldata, "data.pkl")




import config
import measfcts
import os
import numpy as np
import matplotlib.ticker as ticker
import matplotlib

import tenbilac
import momentsml.plot
from momentsml.tools.feature import Feature
import matplotlib.pyplot as plt

import logging
logger = logging.getLogger(__name__)


momentsml.plot.figures.set_fancy(14)

mode = "s"
#mode = "w"
netid = 0 # id of the network to show


if mode is "s":
	workbasedir = os.path.join(config.traindir, config.datasets["tp"])
	conflist = config.shearconflist
elif mode is "w":
	workbasedir = os.path.join(config.traindir, "{}_with_{}".format(config.datasets["tw"], config.datasets["ts"]))
	conflist = config.weightconflist



for (dataconfpath, toolconfpath) in conflist:

	toolconfig = momentsml.learn.tenbilacrun.readconfig(toolconfpath)
	dataconfig = momentsml.learn.tenbilacrun.readconfig(dataconfpath)
	confname = dataconfig.get("setup", "name") + "_" + toolconfig.get("setup", "name")
	trainworkdir = os.path.join(workbasedir, confname) # We will pass this to Tenbilac
	
	
	tblconfiglist = [("setup", "workdir", trainworkdir)]
	ten = tenbilac.com.Tenbilac(toolconfpath, tblconfiglist)
       
	ten._readmembers()
	
	
	net = ten.committee[netid].net
	
	#fig = plt.figure(figsize=(6, 6))
	#ax = fig.add_subplot(1, 1, 1)

	if mode is "s":
		net.onames[0] = r"\hat{g}"+net.onames[0][-1]
		filepath = os.path.join(config.valdir, config.datasets["ts"] + "_" + confname + "_netviz.pdf")
	elif mode is "w":
		net.onames[0] = r"\hat{w}"+net.onames[0][-1]
		filepath = os.path.join(config.valdir,  "{}_with_{}".format(config.datasets["tw"], config.datasets["ts"]) + "_" + confname + "_netviz.pdf")
	
	tenbilac.plot.netviz(ten.committee[netid].net, filepath=filepath)       
	
	

	#momentsml.plot.figures.savefig(os.path.join(config.valdir, config.datasets["ts"] + "_" + confname + "_msbevo"), fig, fancy=True, pdf_transparence=True)
	#plt.show()



"""
Compute derived quantities from adamom measurements.
"""

import astropy.table
import numpy as np
import copy

import logging
logger = logging.getLogger(__name__)

def measfct(catalog, prefix="", **kwargs):
	"""
	We compute 3 new columns:
	- adamom_logflux
	- adamom_g
	- adamom_theta
	"""
	
	cols = ["adamom_flux"]
	for col in cols:
		if col not in catalog.colnames:
			raise RuntimeError("I need column {}".format(col))
			
	output = astropy.table.Table(copy.deepcopy(catalog), masked=True)
	
	output["adamom_logflux"] = np.ma.log10(output["adamom_flux"]) # Negative values get masked.
	
	output["adamom_g"] = np.ma.hypot(output["adamom_g1"], output["adamom_g2"])
	output["adamom_theta"] = 0.5 * np.arctan2(output["adamom_g2"], output["adamom_g1"])
	
	return output
	
	
	
	
	



"""
A little helper to collect checkplots into one single directory, for easier visualization
"""

import config
import os
import shutil


import logging
logging.basicConfig(format=config.loggerformat, level=logging.INFO)
logger = logging.getLogger(__name__)



destdirname = "summary_plots"
destdirpath = config.great3.path(destdirname)
if not os.path.isdir(destdirpath):
	os.makedirs(destdirpath)


for subfield in config.great3.subfields:
	
	logger.info("Working on subfield {}".format(subfield))
	
	#origpath = config.great3.subpath(subfield, "ml", "tw-200c-1000r", "ada4s2w_sum33wmassshort_summary.png")
	#newname = "s2_{}.png".format(subfield)
	
	origpath = config.great3.subpath(subfield, "ml", "tw-200c-1000r", "ada4s1w_sum33wmassshort_summary.png")
	newname = "s1_{}.png".format(subfield)
	
	
	destpath = os.path.join(destdirpath, newname)
	
	shutil.copy(origpath, destpath)
	
	



"""
Modules for metrics computations, using astropy tables as input.
"""
import numpy as np

import calc
import table
import feature

try:
	import statsmodels.api as sm
except:
	pass

from scipy.optimize import curve_fit

import logging
logger = logging.getLogger(__name__)


def metrics(catalog, labelfeature, predlabelfeature, pre_is_res=False):
	"""
	Returns a dict with some simple standard metrics compareing a "label"-column (truth) to the corresponding predictions.
	This function explicitly takes care of masked columns, only unmasked data will be used and also handles 2D columns,
	via the fact that Feature objects have to be given.
	
	:param catalog: an astropy table containing both label and predlabel. It can be masked.
	:param labelfeature: Feature object describing the label
	:param predlabelfeature: idem for predlabel (set Feature.rea as you feel appropriate, if 2D)
	
	:param pre_is_res: if True, predlabelfeature is interpreted as being residues, and
		labelfeature will be added to these before computing the metrics.
	
	:returns: a dict containing
		
		- **predfrac**: Fraction of rows having an unmasked predlabel
		- **rmsd**: the RMSD
		- **m**: multiplicative bias (slope minus one)
		- **merr**: error on multiplicative bias
		- **c**: additive bias
		- **cerr**: error on additive bias

	"""
	
	logger.debug("Computing metrics for label = %s and predlabel = %s" % (labelfeature.colname, predlabelfeature.colname))
	
	#features = [feature.Feature(colname=label, rea=None), feature.Feature(colname=predlabel, rea="full")]

	metcat = feature.get1Ddata(catalog, [labelfeature, predlabelfeature], keepmasked=False)
	
	lab = metcat[labelfeature.colname]
	pre = metcat[predlabelfeature.colname]
	if pre_is_res:
		pre += lab
	
	assert lab.ndim == 1
	assert pre.ndim == 1
	
	# And we prepare the return dict:
	ret = {"predfrac":float(len(metcat))/float(len(catalog))}
	
	ret["rmsd"] = calc.rmsd(lab, pre)
	ret.update(calc.linreg(lab, pre))
	
	return ret


def metricsw(cat, trufeat, predfeat, wfeat=None, sigma_shape=0.2):
	"""
	Same as metrics, but with weights.
	
	sigma = sigma_shape / sqrt(weights)
	
	Bright large galaxies get a weight of 1.0
	All their shear noise is shape noise.
	Shape noise has a sigma of 0.2 -> sigma_shape is 0.2
	
	
	"""
	logger.info("Computing metrics for {} -> {}".format(predfeat.colname, trufeat.colname))
	
	if wfeat is None:
		logger.info("WARNING: not using any weights!")
		metcat = feature.get1Ddata(cat, [trufeat, predfeat], keepmasked=False)
		x = metcat[trufeat.colname]
		y = metcat[predfeat.colname]
		w = np.ones(len(x))
		
	else:
		logger.info("Using {} as weights.".format(wfeat.colname))
		metcat = feature.get1Ddata(cat, [trufeat, predfeat, wfeat], keepmasked=False)
		x = metcat[trufeat.colname]
		y = metcat[predfeat.colname]
		w = metcat[wfeat.colname]
	
	sigma = sigma_shape * np.sqrt(1.0/np.clip(w, 1e-18, 1e18)) # We don't want weights of exactly zero here.

	def f(x, a, b):
		return a * x + b

	p0 = [1.0, 0.0] # initial parameter estimate
	popt, pcov = curve_fit(f, x, y, p0, sigma, absolute_sigma=True)
	perr = np.sqrt(np.diag(pcov))
	
	m = popt[0] - 1.0
	c = popt[1]
	merr = perr[0]
	cerr = perr[1]
	ret = {"m":m, "c":c, "merr":merr, "cerr":cerr}
	
	txt = "m*1e3: %.1f +/- %.1f   c*1e3: %.1f +/- %.1f" % (m*1000.0, merr*1000.0, c*1000.0, cerr*1000.0)
	logger.info("Regression: {}".format(txt))	
	
	return ret
	
	
def linregress_with_errors(x, y, y_err):
    """
	Test for comparision, GSL's gsl_fit_wlinear function
	"""
    
    y_weights = y_err**-2
    total_weight = y_weights.sum()
    
    x_weighted_mean = np.average(x,weights=y_weights)
    y_weighted_mean = np.average(y,weights=y_weights)
    
    dx = x-x_weighted_mean
    dy = y-y_weighted_mean
    
    dx2_weighted_mean = np.average(dx**2,weights=y_weights)
    dxdy_weighted_mean = np.average(dx*dy,weights=y_weights)
    
    slope = dxdy_weighted_mean / dx2_weighted_mean
    intercept = y_weighted_mean - x_weighted_mean*slope
    
    slope_err = np.sqrt(1./(total_weight*dx2_weighted_mean))
    intercept_err = np.sqrt((1.0 + x_weighted_mean**2 / dx2_weighted_mean) / total_weight)
    slope_intercept_covar = -x_weighted_mean / (total_weight*dx2_weighted_mean)

    return slope, intercept, slope_err, intercept_err, slope_intercept_covar



#def wmetrics(cat, trufeat, predfeat, wfeat=None):
#	"""
#	An attempt to get metrics with errors when one has weights, *without* first "bining" things.
#	Warning, this is experimental, not yet fully tested.
#	
#	Note that the scale of weights matters for the errors!
#	Provide weights corresponding to 1/sigma**2
#	"""
#	logger.info("Computing metrics for {} -> {}".format(predfeat.colname, trufeat.colname))
#	
#	if wfeat is None:
#		logger.info("WARNING: not using any weights!")
#		metcat = feature.get1Ddata(cat, [trufeat, predfeat], keepmasked=False)
#		x = metcat[trufeat.colname]
#		y = metcat[predfeat.colname]
#		w = np.ones(len(x))
#		
#	else:
#		logger.info("Using {} as weights.".format(wfeat.colname))
#		metcat = feature.get1Ddata(cat, [trufeat, predfeat, wfeat], keepmasked=False)
#		x = metcat[trufeat.colname]
#		y = metcat[predfeat.colname]
#		w = metcat[wfeat.colname]
#	
#
#	y_err = np.sqrt(1.0/np.clip(w, 1e-18, 1e18)) # We don't want weights of exactly zero here.
#	
#	########### GSL
#	(slope, intercept, slope_err, intercept_err, slope_intercept_covar) = linregress_with_errors(x, y, y_err)
#	txt = "m*1e3: %.1f +/- %.1f   c*1e3: %.1f +/- %.1f" % ((slope-1.0)*1000.0, slope_err*1000.0, intercept*1000.0, intercept_err*1000.0)
#	logger.info("GSL regression: {}".format(txt))	
#	
#
#	########### Scipy
#	def f(x, a, b):
#		return a * x + b
#
#	p0 = [0, 0] # initial parameter estimate
#	popt, pcov = curve_fit(f, x, y, p0, y_err, absolute_sigma=True)
#	perr = np.sqrt(np.diag(pcov))
#	txt = "m*1e3: %.1f +/- %.1f   c*1e3: %.1f +/- %.1f" % ((popt[0]-1.0)*1000.0, perr[0]*1000.0, popt[1]*1000.0, perr[1]*1000.0)
#	logger.info("Scipy regression: {}".format(txt))	
#	
#	
#	########### statsmodels
#	X = sm.add_constant(x)
#	#model = sm.OLS(y, X)
#	model = sm.WLS(y, X, weights=1.0/(y_err**2)) # Yes, this "w" should be proportional to the inverse of the variance
#	
#	results = model.fit(cov_type='fixed scale')
#	print(results.summary())
#	
#	print('Parameters: ', results.params)
#	print('Standard errors: ', results.bse)
#	
#	c = results.params[0]
#	m = results.params[1] - 1.0
#	cerr = results.bse[0]
#	merr = results.bse[1]
#	ret = {"m":m, "c":c, "merr":merr, "cerr":cerr}
#	
#	txt = "m*1e3: %.1f +/- %.1f   c*1e3: %.1f +/- %.1f" % (m*1000.0, merr*1000.0, c*1000.0, cerr*1000.0)
#	logger.info("StatsModels Regression: {}".format(txt))	
#	
#	return ret
#	
#	
	
	
	
	
